//***********************************************************************************************************//
//***********************************************************************************************************//
//                                                                                                           //
// Gleitkomma.cpp and Gleitkomma.h : Used for definition of the class real. An Object of type real           //
//      represents a floating point number with adjustable precision (number of mantissa bits/words).         //
//                                                                                                           //
// Autor: Florian Knigstein, mail@virgusta.eu                                                               //
//                                                                                                           //
// Compiler: avr-gcc                                                                                         //
//                                                                                                           //
// Copyright:                                                                                                //
//      You may use this program free of charge for any purpose you wish provided you agree to the following //
//      License:                                                                                             //
//      Warranty of Provenance and Disclaimer of Warranty. Licensor warrants that the copyright              //
//      in and to the Original Work and the patent rights granted herein by Licensor are owned by the        //
//      Licensor or are sublicensed to You under the terms of this License with the permission of the        //
//      contributor(s) of those copyrights and patent rights. Except as expressly stated in the immediately  //
//      preceding sentence, the Original Work is provided under this License on an "AS IS" BASIS and WITHOUT //
//      WARRANTY, either express or implied, including, without limitation, the warranties of                //
//      non-infringement, merchantability or fitness for a particular purpose. THE ENTIRE RISK AS TO THE     //
//      QUALITY OF THE ORIGINAL WORK IS WITH YOU. This DISCLAIMER OF WARRANTY constitutes an essential part  //
//      of this License. No license to the Original Work is granted by this License except under this        //
//      disclaimer.                                                                                          //
//                                                                                                           //
//      Limitation of Liability. Under no circumstances and under no legal theory, whether in tort           //
//      (including negligence), contract, or otherwise, shall the Licensor be liable to anyone for any       //
//      indirect, special, incidental, or consequential damages of any character arising as a result of      //
//      this License or the use of the Original Work including, without limitation, damages for loss of      //
//      goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages    //
//      or losses. This limitation of liability shall not apply to the extent applicable law prohibits such  //
//      limitation.                                                                                          //
//                                                                                                           //
//***********************************************************************************************************//
//***********************************************************************************************************//


//***** ATTENTION ******************************************************************************************//
// This library makes use of dynamic memory allocation. But malloc's return value is NOT verified to be     //
// non zero, i.e. "out of memory" will not be handled properly. The program will likely crash if out of     //
// memory occurs. For workaround you might modify the program and generate a NaN (not an number) if malloc  //
// returns zero.                                                                                            //
//**********************************************************************************************************//

#include "Gleitkomma.h"

int8 GleitkommaRunden = 0;
		// Wenn GleitkommaRunden > 0 gilt, so wird bei Bedarf immer aufgerundet.
		// Bei negativen Zahlen verkleinert sich in einem solchen Fall der Betrag.
		// Jede der vier Grundrechenarten sowie der Copy-Konstruktor mit Verringerung
		// der Genauigkeit liefern niemals eine Zahl, die kleiner ist als die
		// genaue Zahl, die sich ergeben wrde, wenn man die Eingangsdaten als genau
		// betrachtet. Entsprechendes gilt fr die Umwandlung von Zahlen in der
		// Dezimaldarstellung (Strings) in Gleitkommazahlen sowie fr die Umwandlung
		// in umgekehrter Richtung.
		// Wenn GleitkommaRunden < 0 gilt, so wird bei Bedarf abgerundet.
		// Bei negativen Zahlen vergrert sich in einem solchen Fall der Betrag.
		// Wenn GleitkommaRunden==0, so wird auf die nchste darstellbare Zahl
		// gerundet.

int8 StdZahlMantissewoerter = ((int8)((208L*(long)ANZAHL_DEZIMALE_MANTISSESTELLEN+999L)/1000L));


inline void *operator new(unsigned int anz_bytes)
{
	return malloc(anz_bytes);
}

inline void *operator new[](unsigned int anz_bytes)
{
	return malloc(anz_bytes);
}

inline void operator delete(void *address)
{
	free(address);
}

inline void operator delete[](void *address)
{
	free(address);
}


	// Die Definitionen der folgenden Operatoren erfolgt unten, die Deklarationen
	// mssen hier sein, weil sie vorher bentigt werden.
bool operator>=(const real& x, const real& y);
bool operator<=(const real& x, const real& y);

int16 real::min_exp2 = -(1<<13);
int16 real::max_exp2 =  (1<<13);

real& real::ResizeMantisseAnz16Bits(int8 anz)
		// Diese Funktion darf nur dann innerhalb eines Konstruktors von
		// class GleitkommmaZahl aufgerufen werden, wenn die real
		// schon eine gltige Initialisierung besitzt.
{
	if(anz < 1) anz=1;
	int i;

	if(anz > anz_mts)
	{
		if(anz > AnzMantisseMemory)
		{
			unsigned short *mem=new unsigned short[anz];
			if(false==null && false==unendlich)
			{
				for(i=0; i<anz_mts; i++) mem[i]=mts[i];
				for( ; i<anz ; i++) mem[i]=0;
			}
			delete[] mts;
			anz_mts=AnzMantisseMemory=anz;
			mts=mem;
		}
		else
		{
			if(false==null && false==unendlich)
				for(i=anz_mts; i<anz; i++) mts[i]=0;
			anz_mts=anz;
		}
	}
	else if(anz<anz_mts)
	{
		int alt_anz=anz_mts;
		anz_mts=anz;
		if(false==null && false==unendlich)
		{
			if(GleitkommaRunden==0)
					// zur nchtgelegenen Zahl runden
			{
				if(mts[anz] >= 32768u)
					naechsthoehererBetrag();
			}
			else if((!negativ && GleitkommaRunden > 0) ||
						(negativ && GleitkommaRunden < 0)
					 )
			{
				for(i=anz ; i<alt_anz ; i++)
					if(mts[i] != 0)
					{
						naechsthoehererBetrag();
						break;
					}
			}
		}
	}
	return *this;
}

void real::init(int8 anz_mts_16bits)
		// Nur zum Aufruf innerhalb eines Konstruktors bestimmt.
		// init() erzeugt eine gltige, d.h. korrekt initialisierte real
		// mit dem Wert Null. Fr eine bereits initialisierte real
		// darf NICHT erneut init() aufgerufen werden, weil der alte Speicher fr
		// 'mts' nicht zuvor freigegeben wird.
{
	null=true;
	negativ=false;
	unendlich=false;
	exp2=0;
	if( (anz_mts=anz_mts_16bits) < 1)
		anz_mts=1;
	AnzMantisseMemory=anz_mts;
	mts=new unsigned short[AnzMantisseMemory];	// Der Speicher mu nicht initialisiert
		// werden, weil null==true gesetzt wurde.
}

real::real(const real& x)
		// Copy-Konstruktor, bernahme der Genauigkeit des Parameters
{
	negativ=x.negativ;
	exp2=x.exp2;
	unendlich=x.unendlich;
	null=x.null;
	AnzMantisseMemory=anz_mts=x.anz_mts;
	mts=new unsigned short[AnzMantisseMemory];
	if(false==x.null && false==x.unendlich)
		for(uint8 i=0; i<anz_mts; i++)
			mts[i]=x.mts[i];
}

/*real::real(int16 Exp2, int8 Anz_mts, unsigned short *Mantisse,
								 bool Negativ)
		// Default ist negativ==false.
		// Falls sich durch exp2 oder zusammen mit der Normalisierung der
		// Mantisse ein exp2 < min_exp2 ergibt, so wird NICHT gerundet, sondern
		// die Zahl Null erzeugt.
{
	init(Anz_mts < 1 ? 1 : Anz_mts);

	negativ=Negativ;

	if(Anz_mts < 1 || !Mantisse)
		null=true;
	else
	{
		exp2=Exp2;	// Zu kleiner und zu groer Exponent werden in
			// normalisiere() behandelt.
		for(uint8 i=0; i<Anz_mts; i++)
			mts[i]=Mantisse[i];
		null=false;		// kann von normalisiere() noch gendert werden.
		normalisiere();
	}
}*/

real real::MinMaxPositiveZahl(Anz16bits anz, bool max)
{
	if(anz.anz < 1) anz.anz=1;
	real x(Anz16bits(anz.anz));
	x.mts[0]=max ? 65535u : 32768u;
	for(uint8 i=1; i<anz.anz; i++)
		x.mts[i]=max ? 65535u : 0;
	x.exp2=max ? max_exp2 : min_exp2;
	x.negativ=false;
	x.null=false;
	x.unendlich=false;
	return x;
}

void real::initialize_with_long(const long N, Anz16bits anz)
					// Default fr 'anz' ist 'StdZahlMantissewoerter'
		// Wenn anz.anz < 1 ist, so wird der Wert 1 oder 2 (je nach 'N') genommen.
		// Wenn anz.anz==1 ist, so wird genau eine Mantissestelle (16 bits) nur dann
		// genommen, wenn
		//			-65535u <= N <= 65535u ist.  Sonst werden 2 Mantissestellen genommen.
		// Wenn anz.anz >= 2 ist, so werden anz.anz Mantissestellen (zu je 16 bits)
		// genommen.
		// N mu kleiner sein als 65536L*65536L ; dies ist immer der Fall, wenn
		// long-Werte durch 32 Bits reprsentiert werden.
{
	int16 mind2=anz.anz >= 2 ? anz.anz : 2;
	init(mind2);
	long N0;
	if((N0=N) < 0) { N0=-N; negativ=true; } else negativ=false;

	if(N0 != 0)	// sonst ist schon null==true .
	{
		for(uint8 i=0; i<mind2; i++)
			mts[i]=0;	// ist noch nicht erfolgt, weil vorher null==true.
		mts[1]=(unsigned short)(N0 & 65535L);
		N0 >>= 16;
		mts[0]=(unsigned short)(N0 & 65535L);
		N0 >>= 16;
		exp2=31;
		null=false;
		normalisiere();
	}
	if(anz.anz < 2 && mts[1]==0)
		ResizeMantisseAnz16Bits(1);
}

void real::initialize_with_ulong(const unsigned long N, Anz16bits anz)
{
	int16 mind2=anz.anz >= 2 ? anz.anz : 2;
	init(mind2);
	negativ=false;

	if(0==N)
		null=true; 
	else
	{
		unsigned long N0=N;
		for(uint8 i=0; i<mind2; i++)
			mts[i]=0;	// ist noch nicht erfolgt, weil vorher null==true.
		mts[1]=(unsigned short)(N0 & 65535L);
		N0 >>= 16;
		mts[0]=(unsigned short)(N0 & 65535L);
		N0 >>= 16;
		exp2=31;
		null=false;
		normalisiere();
	}
	if(anz.anz < 2 && mts[1]==0)
		ResizeMantisseAnz16Bits(1);
}

/*real::real(const unsigned long *digits, long __anz_digits, long __bits_per_digit, long __exp2,
								Anz16bits anz)
	// Achtung: digits[0] ist das niederwertigste Wort und digits[__anz_digits-1] das hchstwertigste
	// -- im Gegensatz zur Organisation der Worte in real::mts[]
{
	int imax=(__anz_digits*__bits_per_digit + (bits_per_digit-1)) / bits_per_digit;
	init(anz.anz >= imax ? anz.anz : imax);
	negativ=unendlich=null=false;
	exp2=__exp2 + bits_per_digit*imax-1;

	unsigned long long carry=0;
	long anzb=0, i, j=anz_mts;
	for(i=0; i<__anz_digits && j>0; i++)
	{
		carry += ((unsigned long long)digits[i]) << anzb;
		if((anzb+=__bits_per_digit) >= bits_per_digit)
		{
			anzb -= bits_per_digit;
			mts[--j]=(digit)(carry & ((1L<<(bits_per_digit-1)) + ((1L<<(bits_per_digit-1))-1)));
			carry >>= bits_per_digit;
		}
	}
	while(0!=carry && j>0)
	{
		mts[--j]=(digit)(carry & ((1L<<(bits_per_digit-1)) + ((1L<<(bits_per_digit-1))-1)));
		carry >>= bits_per_digit;
	}
	normalisiere();
}*/

/*real::real(const double d, Anz16bits anz)
					// Default fr 'anz' ist 'StdZahlMantissewoerter'
		// Falls die Mantissenlnge 'anz.anz' so kurz ist, da nicht alle
		// Bits der Mantisse der double-Zahl 'd' aufgenommen werden knnen, so
		// wird unter Beachtung von 'GleitkommaRunden' gerundet.
{
	long bits=std::numeric_limits<double>::digits;
	long AnzMtsD=(bits+15)/16;
	if(AnzMtsD < 1) AnzMtsD=1;
	if(anz.anz < 1) anz.anz=1;
	init(anz.anz > AnzMtsD ? anz.anz : AnzMtsD);		// Null

	if(0==_finite(d))
	{
		if(d==std::numeric_limits<double>::infinity())
		{
			unendlich=true; null=false;
			negativ=false;
		}
		else if(d==-std::numeric_limits<double>::infinity())
		{
			unendlich=true; null=false;
			negativ=true;
		}
	}
	else if(_isnan(d))
	{
		unendlich=null=true;	// NaN
		negativ=false;
	}
	else if(d != 0.0)
	{
		int EXP2;
		double d0=frexp(d > 0 ? d : -d, &EXP2), z;
		exp2=EXP2-1;
		int i;
		for(i=0; i<AnzMtsD; i++)
		{
			d0 *= 65536.0;
			d0 = modf(d0, &z);
			mts[i]=(unsigned short)z;
		}
		while(i < anz_mts)
			mts[i++]=0;
		if(mts[0] < 32768u)
			mts[0]=32768u;		// Sollte niemals vorkommen !
		null=false;
		if(d < 0)
			negativ=true;
		else
			negativ=false;
	}
	if(anz.anz < AnzMtsD)
		ResizeMantisseAnz16Bits(anz.anz);	// ggf. runden gem 'GleitkommaRunden'.
}*/

int16 real::GetLsbNumber() const
		// Wenn die real gleich Null oder Unendlich ist, so wird Null geliefert.
		// Sonst wird die Nummer des niedrigstwertigsten EINS-Bits in der Mantisse
		// der Zahl geliefert. Das hchstwertigste Mantissebit (welches bei einer
		// gltigen, von Null verschiedenen real immer gleich Eins ist),
		// hat die Nummer Null, das niedrigstwertigste Bit hat die Nummer
		//   16*(*this).anz_mts - 1  .
{
	if(null || unendlich || mts[0] < 32768u)  // im letzten Fall: UNGLTIGE real !!!
		return 0;

	uint8 i;
	int16 lsb;
	unsigned short z;
	for(i=anz_mts-1 ; i>0 && 0==mts[i] ; --i)
		;
	z=mts[i]; // Hier ist immer i>=0 .
	lsb=16*i + 15;
	for(unsigned short maske=1 ; maske<32768u && 0==(z & maske) ; )
	{
		maske <<= 1;
		--lsb;
	}
	return lsb;
}

real to_integer(const real& x)
			// Liefert die ganze real, die aus x durch Abschneiden der
			// Kommastellen entsteht.
{
	if(x.unendlich)
		return x;
	if(x.null || x.exp2 < 0)
		return real(real::Anz16bits(x.anz_mts));		// Null.

	real integer = x;
	int16 i=integer.exp2 / 16;	// Es ist i>=0 wegen obiger Abfrage
	if(i < integer.anz_mts)
	{
		integer.mts[i] &= (  (65535u << (15-(x.exp2 & 15))) & 65535u );
		while(++i < integer.anz_mts)
			integer.mts[i]=0;
		// Die Zahl  integer  ist weiterhin normalisiert.
		// Auerdem kann hier fr  integer  nicht der Wert Null entstanden sein,
		// denn der Fall  x.exp2 < 0  wurde schon oben behandelt.
	}
	return integer;
}

void real::naechstkleinererBetrag()
		// Diese Funktion beachtet die Variable 'GleitkommaRunden'
		// nicht.
		// Wenn null==true gilt, wird nichts gemacht.
		// Wenn *this gleich +Unendlich oder -Unendlich ist,
		// wird das Vorzeichen beihehalten und die grte darstellbare
		// positive reelle Zahl bzw. die kleinste darstellbare negative
		// reelle Zahl erzeugt.
		// Wenn *this reell ist, wird das Vorzeichen beibehalten und der
		// nchstkleinere Betrag eingestellt. Wenn jedoch vor dem Aufruf
		// dieser Funktion bereits der kleinst-
		// mgliche Betrag (mit kleinstmglichem Exponenten 'exp2') vorliegt,
		// so wird UNABHNGIG von der Variablen GleitkommaRunden die Zahl Null
		// erzeugt.
{
#ifdef ONLY_NAN_NO_INFINITY
	if(null || unendlich) return;
#else
	if(null) return;
	if(unendlich)
	{
		if(false==negativ)
			*this=MaxPositiveReelleZahl(Anz16bits(anz_mts));
		else
			*this=-MaxPositiveReelleZahl(Anz16bits(anz_mts));
		return;
	}
#endif

	int8 i;
	for(i=anz_mts ; (i--)>0 ; )
		if(mts[i]>0)
			{ --mts[i]; return; }
		else
			mts[i]=65535u;

	if(mts[0] >= 32768u) return;

		// Der folgende Programmteil wird nur dann ausgefhrt, wenn VORHER das
		// erste Mantissewort auf 32768u und alle folgenden auf 0 waren.

	if(exp2<=min_exp2)
		null=true;

	--exp2;
	for(i=0; i<anz_mts; i++) mts[i]=65535u;
}

long to_long(const real& x)
		// Die Nachkommastellen von x werden nicht beachtet, d.h. es wird z.B.
		//  4.1  zu  (long)4  und  -4.1  zu  (long)(-4) .
		// --- ACHTUNG ---: Wenn der so entstehende ganzzahlige Teil von x
		// auerhalb des Bereichs ist, der durch long-Zahlen dargestellt
		// werden kann, so ist das Ergebnis undefiniert !
{
	if(x.null || x.exp2 < 0 || x.unendlich) return 0;
	unsigned long absx;
	int16 i=x.exp2 / 16;	// Es ist i>=0 wegen obiger Abfrage.

	if(i > 3) return 0;	// Dies wird nur im Falle eines Fehlers ausgefhrt.
		// In diesem Fall ist dann die Zahl sicher betragsmig grer als der
		// grte Betrag, der mit einer long-Zahl dargestellt werden kann.
		// Die Abfrage findet nur dazu statt, um in diesem Fehlerfall eine
		// mglicherweise sehr lange Rechenzeit in der folgenden Schleife zu
		// verhindern.

	if(i >= x.anz_mts)
		absx=0;
	else
		absx = ( x.mts[i] >> (15 - (x.exp2 & 15)) );

	long shift=(x.exp2 & 15) + 1;
	while(--i >= 0)
	{
		if(i<x.anz_mts) absx += x.mts[i] << shift;
		shift+=16;
	}
	return x.negativ ? (-((long)absx)) : absx;
}

real modf(const real& x, real *integer)
						// Default fr 'integer' ist der Nullzeiger.
		// Es werden eine ganze Zahl z und eine Zahl y berechnet, so da
		// x == z+y  UND  -1 < y < 1  gilt und auerdem sowohl z als auch y
		// das gleiche Vorzeichen haben wie x . Die Zahl y wird zurckgeliefert.
		// Falls der Zeiger  integer != 0  ist, so mu er auf eine real
		// zeigen; dann wird in  *integer  die Zahl z gespeichert.
{
#ifdef ONLY_NAN_NO_INFINITY
	if(x.null || x.exp2 < 0)
	{
		if(integer)
			*integer=real(real::Anz16bits(x.anz_mts));	// Null
		return x;	// 0 <= abs(x) < 1 .
	}
	if(x.unendlich)
	{
		if(integer) *integer=x; // Null oder NaN
		return x;
	}
#else
	if(x.unendlich)
	{
		if(0 != integer)
			*integer = x;
		return real(real::Anz16bits(x.anz_mts)); // Null
	}
	if(x.null || x.exp2 < 0)
	{
		if(integer)
			*integer=real(real::Anz16bits(x.anz_mts));	// Null
		return x;	// 0 <= abs(x) < 1 .
	}
#endif

	int16 i0=x.exp2 / 16;	// Es ist i0>=0 wegen obiger Abfrage
	real gebrochen = x;

	if(i0 < x.anz_mts)
	{
		gebrochen.mts[i0] &= (  65535u >> ((x.exp2 & 15) + 1)  );
		for(int16 i=i0 ; --i>=0 ; )
			gebrochen.mts[i]=0;
		gebrochen.normalisiere();	// Hier kann auch noch gebrochen.null=true
				// gesetzt werden !
	}
	else
		gebrochen.null=true;

	if(integer)
	{
		*integer = x;
		if(i0 < x.anz_mts)
		{
			(*integer).mts[i0] &= (  (65535u << (15-(x.exp2 & 15))) & 65535u );
			for(int16 i=i0 ; ++i < x.anz_mts ; )
				(*integer).mts[i]=0;
			// Die Zahl  *integer  ist weiterhin normalisiert.
			// Auerdem kann hier fr  *integer  nicht der Wert Null entstanden sein,
			// denn der Fall  x.exp2 < 0  wurde schon oben behandelt.
		}
	}

	return gebrochen;
}

/*bool is_integer(const real& x)
		// Liefert true, falls x ganzzahlig ist, sonst false.
{
	if(x.unendlich || x.exp2 < 0)
		return false;
	if(true==x.null)	// Es gilt false==unendlich
		return true;

	long i=x.exp2 / 16;	// Es ist i>=0 wegen obiger Abfrage
	if(i < x.anz_mts)
	{
		if(0 != (x.mts[i] & (  (65535u >> (1+(x.exp2 & 15)) ) )))
			return false;
		while(++i < x.anz_mts)
			if(0 != x.mts[i])
				return false;
	}
	return true;
}*/

/*real floor(const real& x)
		// Liefert die grte ganze real, die
		// kleiner oder gleich x ist.
{
	if(x.unendlich)
		return x;
	if(x.null || (false==x.negativ && x.exp2 < 0))
		return real(real::Anz16bits(x.anz_mts)); // Null.
	if(x.exp2 < 0)	// dann ist x negativ
		return real(-1L, real::Anz16bits(x.anz_mts)); // -1.0

	real integer = x;
	if(is_integer(x))
		return x;
	else if(false==x.negativ)
	{
		long i=integer.exp2 / 16;	// Es ist i>=0 wegen obiger Abfrage
		if(i < integer.anz_mts)
		{
			integer.mts[i] &= (  (65535u << (15-(x.exp2 & 15))) & 65535u );
			while(++i < integer.anz_mts)
				integer.mts[i]=0;
			// Die Zahl  integer  ist weiterhin normalisiert.
			// Auerdem kann hier fr  integer  nicht der Wert Null entstanden sein,
			// denn der Fall  x.exp2 < 0  wurde schon oben behandelt.
		}
		return integer;
	}
	else
	{
		long i=integer.exp2 / 16;	// Es ist i>=0 wegen obiger Abfrage
		if(i < integer.anz_mts)
		{
			integer.mts[i] &= (  (65535u << (15-(x.exp2 & 15))) & 65535u );
			while(++i < integer.anz_mts)
				integer.mts[i]=0;
			// Die Zahl  integer  ist weiterhin normalisiert.
			// Auerdem kann hier fr  integer  nicht der Wert Null entstanden sein,
			// denn der Fall  x.exp2 < 0  wurde schon oben behandelt.
		}
		return integer
				- real(1L, real::Anz16bits(x.anz_mts));
	}
}

real ceil(const real& x)
		// Liefert die kleinste ganze real, die
		// grer oder gleich x ist.
{
	if(x.unendlich)
		return x;
	if(x.null)
		return real(real::Anz16bits(x.anz_mts)); // Null.
	if(x.exp2 < 0)
		if(x.negativ)
			return real(real::Anz16bits(x.anz_mts)); // Null
		else
			return real(1L, real::Anz16bits(x.anz_mts)); // 1.0

	real integer = x;
	if(is_integer(x))
		return x;
	else if(true==x.negativ)
	{
		long i=integer.exp2 / 16;	// Es ist i>=0 wegen obiger Abfrage
		if(i < integer.anz_mts)
		{
			integer.mts[i] &= (  (65535u << (15-(x.exp2 & 15))) & 65535u );
			while(++i < integer.anz_mts)
				integer.mts[i]=0;
			// Die Zahl  integer  ist weiterhin normalisiert.
			// Auerdem kann hier fr  integer  nicht der Wert Null entstanden sein,
			// denn der Fall  x.exp2 < 0  wurde schon oben behandelt.
		}
		return integer;
	}
	else
	{
		long i=integer.exp2 / 16;	// Es ist i>=0 wegen obiger Abfrage
		if(i < integer.anz_mts)
		{
			integer.mts[i] &= (  (65535u << (15-(x.exp2 & 15))) & 65535u );
			while(++i < integer.anz_mts)
				integer.mts[i]=0;
			// Die Zahl  integer  ist weiterhin normalisiert.
			// Auerdem kann hier fr  integer  nicht der Wert Null entstanden sein,
			// denn der Fall  x.exp2 < 0  wurde schon oben behandelt.
		}
		return integer
				+ real(1L, real::Anz16bits(x.anz_mts));
	}
}*/

real& real::mult2hochX(int16 x)
{
	if(true==unendlich)
	{
#ifndef ONLY_NAN_NO_INFINITY
		if(false==null && x<0 &&
				((false==negativ && GleitkommaRunden < 0)
				|| (true==negativ && GleitkommaRunden > 0)))
		{
			if(false==negativ)
				*this=real::MaxPositiveReelleZahl(Anz16bits(anz_mts));
			else
				*this=-real::MaxPositiveReelleZahl(Anz16bits(anz_mts));
		}
#endif
	}
	else if(null==false)
	{
		exp2 += x;
		check_exp();
	}
	return *this;
}

real& real::RoundUpDown(int16 genau_vorher)
		// ZU BEACHTEN: ES MUSS  2 <= genau_vorher <= 16*anz_mts  gelten.
		// Wenn GleitkommaRunden==0 ist, so gibt RoundUpDown *this zurck.
		// Das gleiche gilt, wenn die real gleich Null oder Unendlich ist.
		// Sonst wird angenommen, da vor dem Aufruf von RoundUpDown()
		// die betreffende real (*this) eine Genauigkeit von 
		//    2 hoch -'genau_vorher'
		// besitzt. Dann sorgt RoundUpDown im Falle von GleitkommaRunden > 0
		// durch Addition einer kleinen Zahl zu der real
		// dafr, da das Ergebnis (der Rckgabewert von RoundUpDown())
		// niemals kleiner ist als eine reelle Zahl, die mit obiger
		// Genauigkeit durch die Zahl *this  (*this : vor der Ausfhrung von
		// RoundUpDown()) angenhert werden kann. Entsprechend ist
		// im Falle von GleitkommaRunden < 0 die von RoundUpDown()
		// zurckgelieferte Zahl niemals zu gro; RoundUpDown() subtrahiert
		// dann eine kleine Zahl.
		// Die "kleine Zahl" wird zu
		//    ( 2 hoch (*this).exp2 )  *  ( 2 hoch -(genau_vorher-1) )
		// gewhlt. Man kann zeigen, da durch diese Addition bzw.
		// Subtraktion die entstehende Zahl noch eine Genauigkeit von
		//   2 hoch -(genau_vorher - 2)   zu einer reellen Zahl besitzt, die durch
		// *this vor RoundUpDown() angenhert wurde.
		// Wenn allerdings durch Addition der "kleinen Zahl" bei unverndertem
		// Exponenten der real die Mantisse grer als 2 wrde,
		// so mu der Exponent um eins erhht werden und die Mantisse
		// (die zunchst >=2 wre), durch 2 dividiert werden. Wenn durch
		// diese Bitverschiebung rechts aber ein 1-Bit herausgeschoben wrde,
		// so mu aufgerundet werden, d.h. insgesamt wird eine etwas
		// grere Zahl addiert. Auch in diesem Fall gilt noch, da die
		// entstehende Zahl mit der Genauigkeit von  2 hoch -(genau_vorher - 2)
		// die reelle Zahl annhert, die vor Aufruf von RoundUpDown() mit
		// der Genauigkeit von  2 hoch -genau_vorher  angenhert wurde.
{
	if(false==null && false==unendlich && 0!=GleitkommaRunden)
	{
		if(genau_vorher < 2) genau_vorher=2;
		if(genau_vorher > 16*anz_mts) genau_vorher=16*anz_mts;

		int16 i=(genau_vorher - 1) / 16;	// i<anz_mts wegen obiger Abfrage.
		long addsub=32768u >> ((genau_vorher - 1) & 15);

		if((GleitkommaRunden > 0 && !negativ) || (GleitkommaRunden < 0 && negativ))
		{
			long s;
			if( (s=((long)mts[i]) + addsub) < 65536L )
				mts[i]=(unsigned short)s;
			else
			{
				mts[i]=(unsigned short)(s - 65536L);
				for( ; --i>=0 ; )
					if(mts[i]<65535u)
						{ ++mts[i]; break; }
					else
						mts[i]=0;
				if(i<0)
					UebertragUndRunden(1, true, 0);
			}
		}
		else // (GleitkommaRunden > 0 && negativ) || (GleitkommaRunden < 0 && !negativ)
		{
			long s;
			if( (s=((long)mts[i]) - addsub) >= 0 )
				mts[i]=(unsigned short)s;
			else
			{
				mts[i]=(unsigned short)(s + 65536L);
				for( ; --i>=0 ; )
					if(mts[i]>0)
						{ --mts[i]; break; }
					else
						mts[i]=65535L;
				if(mts[0] < 32768u)
					normalisiere();
			}
		}
	}
		
	return *this;
}

int8 real::AnzBitsBenutztInLong(long x)
		// Ein negatives x wird zuerst mit -1 multipliziert.
		// Dann wird die Gre des kleinstn zusammenhngenden Blocks von Bits
		// in x geliefert, der alle von Null verschiedenen Bits von x enthlt
		// und immer das niederwertigste Bit enthlt.
{
	if(x < 0) x=-x;
	int8 anz=0;
	while(0!=x) { ++anz; x>>=1; }
	return anz;
}

int8 real::AnzMtsBenutzt() const
		// Ein negatives x wird zuerst mit -1 multipliziert.
		// Dann wird die Gre des kleinstn zusammenhngenden Blocks von Bits
		// in x geliefert, der alle von Null verschiedenen Bits von x enthlt
		// und immer das niederwertigste Bit enthlt.
{
	if(null || unendlich)
		return 0;
	for(int8 i=anz_mts ; i>0 ; )
		if(0!=mts[--i])
			return i+1;
	return 0;
}

real real::_10HochX(int16 x, int16 AbsLbGenau, int16 MinAnzMtsBits)
		// Es sind positive und negative Werte fr x erlaubt.
		// Die Variable GleitkommaRunden wird beachtet. Unabhngig von ihrem Wert
		// wird eine Genauigkeit von 2 hoch -(AbsLbGenau) garantiert. Wenn aber
		// z.B. GleitkommaRunden > 0 ist, so wird zustzlich garantiert, da nie
		// ein gegenber dem wahren (reellen) Wert zu kleiner Wert zurckgeliefert wird.
		//
		// --- Achtung ---: Es kann die Ausnahme 'OverflowError' geworfen werden.
		// Bei zu kleinem x kann eine Rundung auf Null erfolgen, sofern berhaupt
		// der kleinste zulssige Exponent 'exp2' unterschritten wird.
		//
		// Die Alternative, eine globale Funktion _10HochX zu definieren,
		// ist problematisch, weil nur der Rckgabewert vom Typ real
		// nicht verwendet wird, um sie von einer anderen Funktionen mit dem gleichen
		// Namen, den gleichen Argumenten und einem anderen Rckgabetyp unter-
		// scheiden zu knnen.
{
	int16 x0;
	if((x0=x) < 0) x0=-x;
	if(AbsLbGenau < 14) AbsLbGenau=14;
	int16 AnzMtsBits, AnzMtsWords;
	AnzMtsBits=2 + AbsLbGenau;
	if(x0 >= 1) AnzMtsBits += AnzBitsBenutztInLong((x0+1)-1);
		// Der Ausdruck  AnzBitsBenutztInLong(z-1)  liefert fr x0>=1 den
		// zur nchsten ganzen Zahl aufgerundeten Zweierlogarithmus von z.
		// Hier wird z=x0+1 gewhlt, weil im Falle x < 0 unten noch 1 / res
		// gebildet werden mu und der mgliche Rundungsfehler bercksichtigt
		// werden mu.

	if(MinAnzMtsBits > AnzMtsBits)
		AnzMtsWords=(MinAnzMtsBits+15) / 16;
	else
		AnzMtsWords=(AnzMtsBits+15) / 16;

	real res(1L, Anz16bits(AnzMtsWords));
	real pot(10L, Anz16bits(AnzMtsWords));

	if(x < 0) GleitkommaRunden = -GleitkommaRunden;
		// Dann darf 'res' z.B. bei (vorher) GleitkommaRunden > 0 nicht zu gro werden,
		// weil unten 1/res gebildet wird.
	
	while(0 != x0)
	{
		if(0 != (x0 & 1))
			res *= pot;
		pot *= pot;
		x0 >>= 1;
	}

	if(x < 0)
	{
		GleitkommaRunden = -GleitkommaRunden;	// ACHTUNG: Dies mu VOR der Division
			// erfolgen !
		res = number_one() / res;
	}
	
	AnzMtsBits=2 + AbsLbGenau;
	if(MinAnzMtsBits > AnzMtsBits)
		AnzMtsBits=MinAnzMtsBits;
	res.ResizeMantisseAnz16Bits((AnzMtsBits+15)/16);
	return res;
}


real::real(const char *wert, Anz16bits anz)
		// Konstruktion der real aus der im Dezimalformat in 'wert'
		// abgespeicherten Zahl.
		// Der String darf "+INF", "-INF", "NaN" sein oder wie folgt:
		// Fhrende Leerzeichen und '\t' werden ignoriert.
		// 1. Zeichen (optional) :  '-' oder '+'
		// danach (optional): Ziffern '0' bis '9' 
		// danach (optional): Zunchst '.', dann Ziffern '0' bis '9'
		// danach (optional):
		//          'E' oder 'e'  fr "10 hoch"
		//          (optional): '-' oder '+'
		//          Ziffern '0' bis '9'
		// Wenn keine dezimale Mantisse angegeben ist, aber das Exponentzeichen 'E',
		// so wird die Mantisse als Eins angenommen, z.B. stellt "E2" die Zahl
		// 1*(10 hoch 2) = 100  und "E" die Zahl  1*(10 hoch 0) = 1  dar. Wenn weder
		// eine dezimale Mantisse noch das Exponentzeichen (Zeichen 'E') angegeben
		// wurde, so wird die Zahl Null erzeugt.
{
	if(0==wert)
	{
		init(anz.anz);
		return;
	}
	
	int16 AddToExp2=0;
	bool zugewiesen=false;

	unsigned char *DezimalMantisse=0;
	if(anz.anz < 1) anz.anz=1;

	bool neg=false, negexp=false;
	const char *w=wert;
	while(*w && (*w==' ' || *w=='\t')) w++;

	if((w[0]=='n' || w[0]=='N') && (w[1]=='a' || w[1]=='A')
				&& (w[2]=='n' || w[2]=='N')
				&& (w[3]==0 || w[3]==' ' || w[3]=='\t'))
	{
		*this=MakeNaN();
		return;
	}

	if(*w=='-')
	{
		neg=true; ++w;
		GleitkommaRunden = -GleitkommaRunden;	// mu am Ende der Funktion rckgngig
			// gemacht werden !
	}
	else if(*w=='+')
		++w;

#ifndef ONLY_NAN_NO_INFINITY
	if((w[0]=='i' || w[0]=='I') && (w[1]=='n' || w[1]=='N')
			&& (w[2]=='f' || w[2]=='F')
			&& (w[3]==0 || w[3]==' ' || w[3]=='\t'))
	{
		if(neg)
		{
			GleitkommaRunden = -GleitkommaRunden;
			*this=MakeMinusInfinity();
			return;
		}
		*this=MakePlusInfinity();
		return;
	}
#endif

	int16 anz_VorNachKomma=0;
	int16 AddToExp10=0;
	const char *w2=w;
	while(*w2>='0' && *w2<='9') { ++w2; ++anz_VorNachKomma; }
	if(*w2=='.') ++w2;
	while(*w2>='0' && *w2<='9') { ++w2; ++anz_VorNachKomma; --AddToExp10; }
		// Die Nachkommastellen werden zunchst wie Vorkommastellen behandelt
		// und dann durch einen entsprechend kleineren Exponenten 'exp10'
		// korrekt bercksichtigt.

	int16 exp10=0;
	bool exponent_angegeben=false;
	if(*w2=='E' || *w2=='e')
	{
		++w2;
		if(*w2=='-') { negexp=true; ++w2; } else if(*w2=='+') ++w2;
		while(*w2>='0' && *w2<='9')
		{
			if(exp10 > (0x7fff-9) / 10L)
			{
				init(1);	// Null
				if(true==negexp)
				{
					if((false==neg && GleitkommaRunden > 0)
							|| (true==neg && GleitkommaRunden < 0))
						*this=MinPositiveZahl(Anz16bits(anz.anz));
				}
				else
				{
#ifdef ENABLE_GLEITKOMMA_EXCEPTIONS
					throw OverflowError();
#else
					unendlich=true;
					null=false;
#endif
				}
				goto ende;
			}
			exp10=10*exp10 + ( *w2 - '0' );
			++w2;
		}
		if(negexp) exp10=-exp10;
		exponent_angegeben=true;
	}
	exp10 += AddToExp10;

	int16 i;
	if(0==anz_VorNachKomma)
	{
		anz_VorNachKomma=1;
		DezimalMantisse=new unsigned char[1];
		DezimalMantisse[0]=0;
		if(exponent_angegeben)
			DezimalMantisse[0]=1;
	}
	else
	{
		DezimalMantisse=new unsigned char[anz_VorNachKomma];
		for(i=0; i<anz_VorNachKomma; i++) DezimalMantisse[i]=0;
		w2=w;
		i=0;
		while(*w2>='0' && *w2<='9') DezimalMantisse[i++] = *(w2++) - '0';
		if(*w2=='.') ++w2;
		while(*w2>='0' && *w2<='9') DezimalMantisse[i++] = *(w2++) - '0';
	}
	// Nun ist anz_VorNachKomma >= 1 .

		// Im folgenden wird berprft, ob sich die ber den Konstruktorparameter
		// 'wert' festgelegte Dezimalzahl EXAKT durch eine real (Binr-
		// system) mit  16*anz.anz  Mantissebits darstellen lt. Wenn dies der
		// Fall ist, so wird sie durch einen speziellen Algorithmus exakt konvertiert.
		// Im anderen Fall wird sie unter Beachtung der Variablen 'GleitkommaRunden'
		// nherungsweise konvertiert.

	AddToExp2=0;
	zugewiesen=false;

	if(exp10 < 0)
	{
			// Zunchst wird berprft, ob die Zahl Null vorliegt, um dann bei
			// einem negativen Exponenten lange Rechenzeiten in der unteren Schleife
			// zu vermeiden.
		for(i=0; i<anz_VorNachKomma && 0 == DezimalMantisse[i]; i++) ;
		
		if(i==anz_VorNachKomma)
			exp10=0;		// exp10=0 zur Erfllung einer folgenden if-Abfrage.
							// Es wird unten die Zahl Null erzeugt.
		else
		{
				// Bei exp10 < 0 kann es sein, da die im Dezimalsystem vorgegebene Zahl
				// 'wert' im Binrsystem unendlich viele Kommastellen aufweist. Nur falls
				// dies nicht zutrifft und wenn andere Bedingungen erfllt sind, kann
				// 'wert' EXAKT in eine real mit der angegebenen Zahl von
				// Mantissestellen konvertiert werden.
			while(exp10 < 0 &&
						( DezimalMantisse[anz_VorNachKomma-1]==0 ||
						  DezimalMantisse[anz_VorNachKomma-1]==5
				  )   )
						// die als ganze Zahl gesehene DezimalMantisse mu durch
						// 5 teilbar sein !
			{
						// Division von 'DezimalMantisse' durch 5, Division durch 2
						// durch Dekrementieren von AddToExp2, Multiplikation mit 10
						// durch Inkrementieren von exp10 ==> es wird weiterhin die
						// gleiche Zahl reprsentiert.
				int8 u=0;
				for(i=0; i<anz_VorNachKomma; i++)
				{
					int8 save=DezimalMantisse[i];
					DezimalMantisse[i] = (char)((10*u + save) / 5);
					u = save % 5;	// 10*u war im vorigen Befehl immer durch 5 teilbar.
				}
						// nun ist u==0, weil die Zahl durch 5 teilbar war.
				--AddToExp2;
				++exp10;
			}
		}
	}

	if(exp10 >= 0)
			// Wenn aber (immer noch) exp10 < 0 gilt, so war die 'DezimalMantisse'
			// in der vorigen  if(exp10 < 0) -Abfrage nicht gengend oft durch 5
			// teilbar; dann kann die Zahl 'wert' nicht exakt konvertiert werden.
	{
			// Nun wird die DezimalMantisse, welche als ganze Zahl interpretiert wird,
			// solange durch 2 geteilt, bis dadurch
			// eine ungerade Zahl entsteht. Durch jeweiliges Inkrementieren von
			// AddToExp2 wird sichergestellt, da immer noch die gleiche Zahl
			// reprsentiert wird.
		int8 vor_nullen=0;
		while(vor_nullen < anz_VorNachKomma  &&  0 == DezimalMantisse[vor_nullen])
			++vor_nullen;

			// Die folgende while-Schleife darf NICHT ausgefhrt werden, wenn
			// alle Mantissestellen gleich Null sind, also wenn
			//  vor_nullen==anz_VorMachKomma  gilt !
		while(vor_nullen < anz_VorNachKomma &&
				0 == (DezimalMantisse[anz_VorNachKomma-1] % 2))
					// die als ganze Zahl gesehene DezimalMantisse mu durch
					// 2 teilbar sein !
		{
			int8 u=0;
			for(i=vor_nullen; i<anz_VorNachKomma; i++)
			{
				int8 save=DezimalMantisse[i];
				DezimalMantisse[i] = (char)((10*u + save) / 2);
				u = save % 2;	// 10*u ist immer durch 2 teilbar.
			}
					// nun ist u==0, weil die Zahl durch 2 teilbar war.
			if(0==DezimalMantisse[vor_nullen])
				++vor_nullen;

			++AddToExp2;
		}

			// Nun liegt mit DezimalMantisse eine ungerade ganze Zahl vor, welche
			// nach Multiplikation mit  '10 hoch exp10'  und mit  '2 hoch AddToExp2'
			// die durch den Konstruktorparameter 'wert' definierte Zahl ergibt.
			// Letztere Zahl ist auch gleich
			//      DezimalMantisse * (5 hoch exp10) * (2 hoch (exp10 + AddToExp2))
			// Um zu testen, ob sie EXAKT durch eine real mit der vorge-
			// gebenen Mantisselnge darstellbar ist, braucht man den letzten Faktor
			// (2 hoch ...) nicht zu bercksichtigen, weil dieser direkt in den
			// Exponenten der binren real eingeht.
			// Weil aber das ganzzahlige Produkt
			//    DezimalMantisse * (5 hoch exp10)
			// aus zwei ungeraden Faktoren besteht und somit ungerade ist, ist in
			// dessen Binrdarstellung die Binrziffer mit der Wertigkeit Eins
			// gleich Eins. Um die Gesamtzahl der Mantissebits zu bestimmen, die
			// fr eine EXAKTE Darstellung dieser Zahl notwendig sind, mu man also
			// den Zweierlogarithmus aus der Wertigkeit des hchstwertigsten Bits
			// bestimmen, das in der ganzen Zahl  DezimalMantisse * (5 hoch exp10)
			// auftritt, und dazu Eins addieren.

		if(vor_nullen == anz_VorNachKomma)
		{
			(*this).init(anz.anz);
			zugewiesen=true;
		}
		else
		{
				// Wichtig: bei anz_VorNachKomma==vor_nullen darf min_anz_bits nicht
				// berechnet werden, daher die obige Abfrage.
			int16 min_anz_bits = (int16)((anz_VorNachKomma-vor_nullen)-1 +
				(int16)(( (long)(exp10+(anz_VorNachKomma-vor_nullen)-1) * 2321L + 990L)/1000L));
							// Die Zahl 2.3219 ist etwas kleiner als  log(5) / log(2) .
			if(16*anz.anz >= min_anz_bits)
			{		// ---ABER---: Nicht diese Bedingung, sondern z.B. die Bedingung
					//   16*anz.anz >= max_anz_bits  wrde die exakte Konvertierbarkeit
					// der Zahl 'wert' garantieren. Ist letztere Bedingung nicht, aber
					// die Bedingung der if-Abfrage erfllt, so kann man nicht auf
					// einfache Weise, sondern nur durch explizites Konvertieren
					// (mit hherer Genauigkeit) feststellen, ob eine Konversion in
					// eine real mit  16*anz.anz  Mantissebits exakt erfolgen
					// kann.
				int16 max_anz_bits = (int16)((anz_VorNachKomma-vor_nullen) +
					(int16)(( (long)(exp10+(anz_VorNachKomma-vor_nullen)) * 2322L + 1010L)/1000L));
								// Die Zahl 2.3220 ist etwas grer als  log(5) / log(2) .

				// Weil unter Umstnden etwas mehr als 'min_anz_bits', nmlich maximal
				// 'max_anz_bits' binre Mantissebits fr eine exakte Darstellung der
				// vorgegebenen Zahl bentigt werden, kann es sein, da
				// die  16*anz.anz  Mantissebits dazu nicht ausreichen. Im folgenden wird
				// zunchst eine exakte Rechnung mit Hilfe von (mindestens) 'max_anz_bits'
				// Mantissenbits durchgefhrt. Das Ergebnis wird dann in eine
				// real mit den geforderten  16*anz.anz  Mantissebits konvertiert.
				// Falls es sich mit dieser Zahl von Mantissenbits nicht exakt darstellen
				// lt, so entsteht dadurch ein Rundungsfehler.

				int16 anz_mts_words=(max_anz_bits + 15) / 16;
				real ExaktZahl=real(Anz16bits(anz_mts_words));
							// ExaktZahl = Null
				for(i=vor_nullen; i<anz_VorNachKomma; i++)
					ExaktZahl = ExaktZahl * number_ten() +
									real((long)DezimalMantisse[i], Anz16bits(1));
				ExaktZahl *= _10HochX(exp10, 16*anz_mts_words);		// Die Zahl
					// '10 hoch exp10' wird IMMER exakt berechnet. Denn '5 hoch exp10'
					// lt sich in  16*anz_mts_words  Mantissebits darstellen, weil gilt:
					//    16*anz_mts_words >= max_anz_bits >= ...
					//    ... max_anz_bits >= floor(exp10 * ln(5)/ln(2)+1.01)
				ExaktZahl.exp2 += AddToExp2;
			
				(*this).init(anz.anz);	// WICHTIG: Die real (*this) mu
						// initialisiert sein, bevor unten der Zuweisungsoperator ( = )
						// auf sie angewandt werden kann.

					// --- An dieser Stelle ist ExaktZahl in jedem Fall eine exakte
					// Konversion der Zahl 'wert'.
				ExaktZahl.ResizeMantisseAnz16Bits(anz.anz);	// eventuell mit RUNDUNGSFEHLER (aber
						// nur, wenn die Zahl 'wert' sich prinzipiell nicht exakt mit
						//  16*anz.anz  binren Mantissebits darstellen lt).
				*this = ExaktZahl;	// NACH init() !!!
				check_exp();
				zugewiesen=true;
					// ACHTUNG: Die Variable (*this).negativ wird unten eventuell noch
					// gesetzt !
			}
		}
	}

	if(false==zugewiesen)
	{
				// Die vorgegebene Dezimalzahl 'wert' lt sich nicht exakt durch
				// eine real mit  16*anz.anz  Mantissebits darstellen.
				// ==> Hier wird nun nherungsweise konvertiert.
		int16 AnzMtsBits, AnzMtsWords;
		AnzMtsBits=anz.anz*16 + 16;	// Weil nur eine entsprechende Genauigkeit
				// garantiert werden kann, wrden anz.anz*16 - 2 Bits ausreichen. Die
				// zustzlichen bewirken, da meistens eine hhere Genauigkeit erreicht
				// wird.
		AnzMtsBits += AnzBitsBenutztInLong((2 + 2*anz_VorNachKomma)-1);
			// Der Ausdruck  AnzBitsBenutztInLong(z-1)  liefert fr z>=1 den
			// zur nchsten ganzen Zahl aufgerundeten Zweierlogarithmus von z.
		AnzMtsWords=(AnzMtsBits+15) / 16;
		real zahl=real(Anz16bits(AnzMtsWords));		// zahl = Null

		for(i=0; i<anz_VorNachKomma; i++)
			zahl=zahl*number_ten() + real((long)DezimalMantisse[i], Anz16bits(1));
		zahl *= _10HochX(exp10, AnzMtsBits);
		zahl.exp2 += AddToExp2;

		(*this).init(anz.anz);	// WICHTIG: Die real (*this) mu
					// initialisiert sein, bevor unten der Zuweisungsoperator ( = )
					// auf sie angewandt werden kann.
		zahl.ResizeMantisseAnz16Bits(anz.anz);		// ... damit in der folgenden Zuweisung kein
					// Freigeben und erneutes Reservieren des Speichers notwendig ist.

		*this=zahl;		// NACH init() !!!
			// Es kann auch  zahl = Null  sein.
		check_exp();
	}

ende:
	if(neg)
	{
		GleitkommaRunden = -GleitkommaRunden;
			// Nun hat GleitkommaRunden den gleichen Wert wie vor Aufruf des Konstruktors.
		negativ=true;	// nach dem Befehl *this=zahl !
	}

	if(0 != DezimalMantisse)
		delete[] DezimalMantisse;
}

char *real::to_decimalExp(int8 anz_dezimal_mantisse, bool MantisseUndExponentGetrennt/*=false*/,
							int16 *ExponentBasis10/*=0*/) const
		// Es werden 'anz_dezimal_mantisse' Dezimalstellen erzeugt, wenn die Zahl
		// von Null verschieden ist.
		// ACHTUNG: anz_dezimal_mantisse darf maximal 32 sein.
{
	if(anz_dezimal_mantisse>32)
		anz_dezimal_mantisse=32;
	static char DezimalString[50];
	DezimalString[0]=0;
	if(is_zero()) { DezimalString[0]='0'; DezimalString[1]=0; return DezimalString; }
	if(is_NaN()) { strcat(DezimalString, "NaN"); return DezimalString; }
#ifndef ONLY_NAN_NO_INFINITY
	if(is_plus_infinity()) { strcat(DezimalString, "+INF"); return DezimalString; }
	if(is_minus_infinity()) { strcat(DezimalString, "-INF"); return DezimalString; }
#endif
	
	if(anz_dezimal_mantisse < 1) anz_dezimal_mantisse=1;

	real konst_1 =number_one();
	real konst_10=number_ten();

	real positiv = *this;
	if(negativ)
	{
		strcat(DezimalString, "-");
		positiv.negativ=false;
		GleitkommaRunden= - GleitkommaRunden;	// unten rckgngig !!!
	}

	int16 i;

		// Im folgenden wird berprft, ob sich die real *this mit der
		// vorgegebenen Zahl von 'anz_dezimal_mantisse' dezimalen Mantissestellen
		// EXAKT darstellen lt. Wenn ja, so wird die exakte Konvertierung ermglicht.

	if(positiv.mts[0] < 32768u) positiv.mts[0]=32768u;	// UNGLTIGE real !!!
		// Sollte niemals vorkommen (der Zugriff auf die Mantisse ist erlaubt, weil
		// der Fall positiv.null==true schon abgefangen wurde).
		// Die Abfrage wurde wegen dem folgenden Aufruf GetLsbNumber() vorgenommen.

	int16 AddToExp10=0, lsb;

	while(positiv.exp2 > (lsb = positiv.GetLsbNumber()))
		// positiv.exp2 - lsb > 0  ==>  'positiv' ist eine ganze, durch zwei teilbare Zahl
		// Um eventuell vorhandene niederwertige Nullen in der Dezimaldarstellung von
		// 'positiv' zu eliminieren (und als Ausgleich 'AddToExp10' zu erhhen), wird
		// die Bitfolge der Mantisse bis zum niederwertigsten 1-Bit als ganze Zahl
		// interpretiert und so lange durch 10 dividiert, bis sie nicht mehr durch 10
		// teilbar ist.
	{
		real quotient=real(Anz16bits(anz_mts));
			// Die Zwischenspeicherung in 'quotient' ist notwendig, weil eventuell
			// die 'positiv' nicht durch 5 teilbar ist und dann nicht verndert werden
			// darf (dann wird unten 'break' ausgefhrt).
		int16 alt_exp2=positiv.exp2;
		positiv.exp2=lsb + 1;
			// 'positiv' ist nun eine ganze Zahl und durch 2, aber nicht durch 4 teilbar.
			// Diese vorbergehende Vernderung des Exponenten ist notwendig,
			// um eine exakte Rechnung im folgenden fmod()-Aufruf zu ermglichen.
		if(false==fmod(positiv, konst_10, &quotient).is_zero())
		{
			positiv.exp2=alt_exp2;
			break;
		}
		quotient.exp2 += alt_exp2 - (lsb+1);
		positiv=quotient;		// Am Beginn der Schleife war 'positiv' durch 10 teilbar,
			// denn die Teilbarkeit durch 2 war die Schleifenbedingung und die
			// Teilbarkeit durch 10 wurde festgestellt. Nun enthlt 'positiv' den
			// 10-ten Teil seinen vorigen Wertes.
		++AddToExp10;
	}

	char *dezimalMtsMem;
	int16 _10Exp;
	int16 maxDezStelle, minDezStelle;
			// maxDezStelle soll gegenber dem wahren Wert nie zu gro berechnet
			// werden, minDezStelle nie zu klein berechnet werden.
			// Diese Variablen werden im folgenden so berechnet, da gilt:
			// Die hchstwertigste Stelle der Dezimaldarstellung von 'positiv' besitzt
			// eine Wertigkeit von (mindestens) 10 hoch maxDezStelle, die
			// niederwertigste Stelle eine Wertigkeit von (hchstens)
			//  10 hoch minDezStelle . Beide Werte 'maxDezStelle' und 'minDezStelle'
			// knnen auch negativ sein.

	if(positiv.exp2 >= 0)
		maxDezStelle=(int16)((((long)positiv.exp2)*3010L)/10000L);
	else
		maxDezStelle=(int16)((((long)positiv.exp2)*3011L-9999L)/10000L);
			// 0.3010 ist etwas kleiner und 0.3011 etwas grer als ln(2) / ln(10) .

		// 'lsb' ist hier gleich GetLsbNumber() .
	if(positiv.exp2 >= lsb)		// ganze Zahl
		minDezStelle=0;
				// Denn: 'positiv' ist wegen obiger Schleife nicht mehr durch 10 teilbar.
	else
		minDezStelle = positiv.exp2 - lsb;	// minDezStelle < 0 .
				// Denn: positiv.exp2 - lsb gibt die Wertigkeit des niederwertigsten
				// Bits der exponentfreien Binrdarstellung von 'positiv' an.
				// Es gilt z.B.  2 hoch -8 == (10 hoch -8) * 5 hoch 8 .

		// Falls die real (*this) exakt darstellbar ist, so ist die
		// Bedingung der folgenden if-Abfrage erfllt. In Einzelfllen kann sie
		// auch erfllt sein, wenn die Zahl nicht exakt darstellbar ist, nmlich
		// deshalb, weil maxDezStelle in manchen Fllen etwas kleiner sein kann
		// als der wahre Wert.
	if(maxDezStelle - minDezStelle + 1 <= anz_dezimal_mantisse)
	{		// --- Exakte Konvertierung ins Dezimalsystem ---

			// Nun wird maxDezStelle erneut so berechnet, da es nie zu klein
			// (statt wie bisher zu gro) gegenber dem wahren Wert ist.
		if(positiv.exp2+1 >= 0)
			maxDezStelle=(int16)((((long)positiv.exp2+1)*3011L)/10000L);
		else
			maxDezStelle=(int16)((((long)positiv.exp2+1)*3010L-9999L)/10000L);

		int16 AnzStellen=maxDezStelle - minDezStelle + 1;
		if(anz_dezimal_mantisse > AnzStellen)
			AnzStellen=anz_dezimal_mantisse;
		dezimalMtsMem=new char[AnzStellen + 2];		// Ein zustzliches
					// 'char' fr den Dezimalpunkt '.' und einen fr das terminierende
					// Nullbyte des C-Strings.
		
		real zwischen=real(Anz16bits(anz_mts));
		real vorKomma=real(Anz16bits(anz_mts));
		real nachKomma=modf(positiv, &vorKomma);

			// Die folgenden Mantisseerweiterungen sind notwendig, damit in den
			// Berechnungen der Vorkommastellen bzw. Nachkommastellen keine
			// Rundungen vorgenommen werden mssen.
		nachKomma.ResizeMantisseAnz16Bits(nachKomma.anz_mts + 1);
		vorKomma.ResizeMantisseAnz16Bits( ((vorKomma.exp2+1) + 15) / 16 );
					// Die Erweiterung von vorKomma ist fr korrektes Arbeiten
					// von fmod() (siehe unten) erforderlich.
					// Es wird vorKomma.exp2+1 verwendet wegen der binren Mantisse, die
					// fast den Wert 2 annehmen kann.

		int16 s=0;
		_10Exp=-1;
		if(false==vorKomma.null)
		{
					// Wenn dieser Programmteil ausgefhrt wird und wenn eine ganze
					// Zahl vorliegt (nachKomma = Null), so ist diese aber nie
					// durch 10 teilbar wegen obiger Schleife zur Beseitigung von
					// Faktoren 10.
			while(false==vorKomma.null && s<AnzStellen)
			{			// Die Bedingung s<AnzStellen ist normal immer erfllt.
				int8 dez=to_long(fmod(vorKomma, konst_10, &zwischen));
				if(dez < 0) dez=0;	// eigentlich unntig.
				if(dez > 9) dez=9;	// eigentlich unntig.
				dezimalMtsMem[1 + (s++)] = (char)('0' + dez);
						// fmod() rechnet exakt. Die dafr notwendige Bedingung ist
						// dadurch erfllt, weil die Mantisselnge von  vorKomma  oben
						// ausreichend erweitert wurde.
				vorKomma = zwischen;
				++_10Exp;
			}
				// Stellen in die richtige Reihenfolge bringen. Man konnte
				// sie deshalb nicht gleich vorher richtig anordnen, weil nicht
				// exakt bekannt war, wieviele Stellen es insgesamt sind.
			for(i=0; i < s-1-i; i++)
			{
				int8 save=dezimalMtsMem[1 + i];
				dezimalMtsMem[1 + i] = dezimalMtsMem[1 + s-1-i];
				dezimalMtsMem[1 + s-1-i] = (char)save;
			}
		}

		nachKomma *= konst_10;
		while(false==nachKomma.null && s<AnzStellen)
		{				// Die Bedingung s<AnzStellen ist normal immer erfllt.
			int8 dez=to_long((const real&)nachKomma);
			nachKomma=konst_10 * modf(nachKomma);

				// Eigentlich kann nun nachKomma nicht auerhalb des Bereichs
				// von je einschlielich 0 bis 9 liegen, denn in
				// dieser Schleife treten keine Rundungsfehler auf.
				// Die folgenden Abfragen sind eigentlich nicht notwendig.
			if(dez < 0) dez=0;
			if(dez > 9) dez=9;

			if(0==s && 0==dez)
				--_10Exp;	// Z.B. 0.0625 ==> 6.25E-2 . Allg.: keine fhrenden Nullen !
			else
				dezimalMtsMem[1 + (s++)] = (char)('0' + dez);
		}

		// An dieser Stelle ist nicht nur die hchstwertigste, sondern auch die
		// niederwertigste nach dezimalMtsMem[] geschriebene Dezimalstelle von Null
		// verschieden, denn wenn 'positiv' vorher z.B. eine ganze Zahl war, so
		// sind Nullen am Ende schon entfernt worden, und statt dessen wurde
		// 'AddToExp10' erhht.

		while(s < anz_dezimal_mantisse)
			dezimalMtsMem[1 + (s++)] = '0';

		bool aufrunden=false;
		if(s > anz_dezimal_mantisse) // wenn ja ==> doch keine exakte Konvertierung mglich
			if(0==GleitkommaRunden)
			{
				if(dezimalMtsMem[1 + anz_dezimal_mantisse] >= '5')
					aufrunden=true;
			}
			else if(GleitkommaRunden > 0)
				aufrunden=true;	// Denn: Es wurden wegen der zweitletzten if-Abfrage
					// mehr Dezimalstellen als 'anz_dezimal_mantisse' berechnet, von
					// denen die letzte von Null verschieden ist (siehe oben).

		if(aufrunden)
		{
						// NICHT bei  i=anz_dezimal_mantisse-1  anfangen, denn
						// der erste Wert dezimalMtsMem[0] ist zunchst unbenutzt.
			for(i=anz_dezimal_mantisse ; i>0 ; --i)
				if(dezimalMtsMem[i] < '9')
				{
					++dezimalMtsMem[i];
					break;
				}
				else
					dezimalMtsMem[i]='0';
			if(0==i)
			{
				++_10Exp;
				dezimalMtsMem[1]='1';
				for(i=2; i<=anz_dezimal_mantisse; i++)
					dezimalMtsMem[i]='0';
			}
		}
	}
	else
	{
			// Im folgenden wird eine nherungsweise Konvertierung ins Dezimalsystem
			// vorgenommen, wobei die Genauigkeit mindestens der gewnschten Zahl
			// 'anz_dezimal_mantisse' von Dezimalstellen entspricht.
		int16 soll_genau=(int16)((3323L * (long)anz_dezimal_mantisse + 2001L)/1000L);
					// 3.3220 ist etwas grer als  ln(10)/ln(2) .
		if(16 * anz_mts > soll_genau)
			soll_genau=16 * (*this).anz_mts;
		_10Exp=0;

		if(positiv < konst_1)
		{
			int16 grob_AbsExp10 = (int16)((((long)(-positiv.exp2+1)) * 31L + 20L)/100L);
				// 0.31 ist etwa gleich, aber ein wenig grer als ln(2) / ln(10) .
			int16 anz_bits=soll_genau + 8 +
						(grob_AbsExp10 <= 7 ? 3 : AnzBitsBenutztInLong(grob_AbsExp10));
			int16 ZwischenAnzMts=(anz_bits + 15)/16;
				// Bei der unten erfolgenden Vernderung von 'positiv' erhlt
				// diese Zahl auch 'ZwischenAnzMts' Mantissewrter zu je 16 Bits.

			real pot10=real(10L, Anz16bits(ZwischenAnzMts)); // nicht number_ten(), weil Anz16bits(ZwischenAnzMts)
					// Es soll immer gelten: pot10 == 10 hoch (1L << bit_nr) .
			int16 bit_nr=0;
			const int16 max_bit_nr=14;
			while(positiv * pot10 < konst_1  &&  bit_nr < max_bit_nr)
			{
				++bit_nr;
				pot10 *= pot10;
			}
			// Wenn z.B. GleitkommaRunden > 0 ist, so ist nun 'pot10' niemals zu klein.

			while(positiv < konst_1 && bit_nr >= 0)
			{
				if(positiv * pot10 < konst_10)
				{
					positiv = positiv * pot10;
					_10Exp += 1u << bit_nr;
				}
				if(bit_nr > 0)
					pot10=sqrt(pot10);	// Wenn z.B. GleitkommaRunden > 0 ist, so ist an
												// dieser Stelle 'pot10' niemals zu klein.
				--bit_nr;
			}
			_10Exp = -_10Exp;
			// Wenn z.B. GleitkommaRunden > 0 ist, so ist nun 'positiv' niemals zu klein,
			// vorausgesetzt da Zahl der Faktoren 10, mit denen 'positiv' multipliziert
			// wurde, als richtig betrachtet wird.


				// Wenn die vorigen Rechnungen mit unendlich hoher Genauigkeit
				// durchgefhrt worden wren, so wre jetzt in jedem Fall
				//     positiv >= 1
				// Tatschlich kann selten auch der Fall vorkommen, da positiv nun
				// geringfgig kleiner ist als eins. In diesem Fall erfolgt unten eine
				// Multiplikation mit 10 und eine Korrektur des Exponenten.
		
			if(positiv < konst_1)
			{		// tritt nur selten auf.
				positiv *= konst_10;		// Nun ist  positiv >= 1 .
					// Weiter ist nun  positiv < 10 , denn bei dieser
					// Multiplikation treten wegen der vorangehenden Mantisseerweiterung
					// von  'positiv'  keine Rundungsfehler auf. Durch die aktuelle
					// Anweisung wird die Mantisse von  'positiv'  lnger, aber wegen
					// der vorangehenden Erweiterung der Mantisse um 16 bits bleiben
					// noch gengend Nullbits am Ende brig, um sicherzustellen, da
					// auch die folgende Schleife ohne Rundungsfehler ausgefhrt wird.
				_10Exp--;
			}
		}
		else
		{
			int16 grob_AbsExp10 = (int16)((((long)(positiv.exp2+1)) * 31L + 20L)/100L);
				// 0.31 ist etwa gleich, aber ein wenig grer als ln(2) / ln(10) .
			int16 anz_bits=soll_genau + 8 +
						(grob_AbsExp10 <= 7 ? 3 : AnzBitsBenutztInLong(grob_AbsExp10));
			int16 ZwischenAnzMts=(anz_bits + 15)/16;
				// Bei der unten erfolgenden Vernderung von 'positiv' erhlt
				// diese Zahl auch 'ZwischenAnzMts' Mantissewrter zu je 16 Bits.

			real pot10=konst_1 / real(10L, Anz16bits(ZwischenAnzMts)); // nicht number_ten(), weil Anz16bits(ZwischenAnzMts)
					// Es soll immer gelten: pot10 == 10 hoch (-(1L << bit_nr)) .
			int16 bit_nr=0;
			const int16 max_bit_nr=14;
			while(positiv * pot10 >= konst_10  &&  bit_nr < max_bit_nr)
			{
				++bit_nr;
				pot10 *= pot10;
			}
			// Wenn z.B. GleitkommaRunden > 0 ist, so ist nun 'pot10' niemals zu klein.

			while(positiv >= konst_10 && bit_nr >= 0)
			{
				if(positiv * pot10 >= konst_1)
				{
					positiv = positiv * pot10;
					_10Exp += 1L << bit_nr;
				}
				if(bit_nr > 0)
					pot10=sqrt(pot10);	// Wenn z.B. GleitkommaRunden > 0 ist, so ist an
												// dieser Stelle 'pot10' niemals zu klein.
				--bit_nr;
			}
			// Wenn z.B. GleitkommaRunden > 0 ist, so ist nun 'positiv' niemals zu klein,
			// vorausgesetzt da Zahl der Faktoren 1.0/10.0, mit denen 'positiv'
			// multipliziert wurde, als richtig betrachtet wird.

				// Wenn die vorigen Rechnungen mit unendlich hoher Genauigkeit
				// durchgefhrt worden wren, so wre jetzt in jedem Fall
				//     positiv < 10
				// Tatschlich kann selten auch der Fall vorkommen, da positiv nun
				// geringfgig grer ist als 10. In diesem Fall wird  'positiv'
				// noch einmal durch 10 dividiert und der Exponent korrigiert. Die
				// dadurch entstehende Zahl nhert
				// dann die tatschliche real mit ausreichender Genauigkeit an.

				// --ACHTUNG-- : Im Gegensatz zum if-Teil der obigen Abfrage
				//    if(positiv < konst_1)
				// mu in diesem else-Teil die Behandlung des Falls 'positiv >= konst_10'
				// VOR der Erweiterung der Mantisse um 16 Bits erfolgen, denn die
				// Division durch 10 wrde nach der Mantisseerweiterung eveltuell alle
				// neuen Null-Bits auffllen, so da in der folgenden Schleife nicht
				// mehr die Abwesenheit von Rundungsfehlern garantiert werden knnte.
			if(positiv >= konst_10)
			{		// tritt nur selten auf.
				positiv /= konst_10;		// Nun ist  positiv < 10 .
					// Weiter ist nun  positiv >= 1 , denn auch im Fall
					//   GleitkommaRunden < 0  ergibt die Division einer Zahl, die
					// nicht kleiner ist als 10, durch 10 eine Zahl, die nicht kleiner
					// ist als 1.

				if(positiv < konst_1)
					positiv=real(1L, Anz16bits(positiv.anz_mts)); // nicht number_one(), weil Anz16bits(positiv.anz_mts)
						// Diese Abfrage ist eigentlich nicht notwendig (siehe den vorigen
						// Kommentar). Wenn diese Bedingung trotzdem erfllt wre, so
						// wrde das Setzen auf 1 im Falle von  'GleitkommaRunden' < 0
						// kein Widerspruch sein, denn vorher war ja  positiv >= konst_10 .

				_10Exp++;
			}
		}

		// Die folgende Erweiterung der Mantisse ist notwendig, um sicherzustellen,
		// da in der gesamten folgenden Schleife keine Rundungsfehler auftreten.
		// Dies ist mglich, weil bei jedem Schleifendurchlauf durch modf()
		// die Vorkommastellen von  'positiv'  entfernt werden. Diese freiwerdenden
		// Bits bewirken, da die nachfolgende Multiplikation mit 10 ohne
		// eine Rundung erfolgen kann. Allerdings ist die nun folgende Mantisse-
		// erweiterung von  'positiv'  notwendig. Denn sonst wrden z.B.
		// bei der Zahl  32767 / 16384  und 16 Bits Mantisse doch ein Rundungs-
		// fehler auftreten, weil beim ersten Schleifendurchlauf weniger Bits
		// vor dem Komma entfernt wrden als nachher durch die Multiplikation
		// mit 10 neu bentigt werden.
		positiv.ResizeMantisseAnz16Bits(positiv.anz_mts + 1);

		dezimalMtsMem=new char[anz_dezimal_mantisse + 2];

		for(i=0; i<anz_dezimal_mantisse; i++)
		{
			int8 vorKomma=to_long((const real&)positiv);

			positiv=konst_10 * modf(positiv);

				// Eigentlich kann vorKomma nicht auerhalb des Bereichs
				// von je einschlielich 0 bis 9 liegen, denn in
				// dieser Schleife treten keine Rundungsfehler auf.
				// Die folgenden Abfragen sind eigentlich nicht notwendig.
			if(vorKomma < 0) vorKomma=0;
			if(vorKomma > 9) vorKomma=9;

			dezimalMtsMem[i+1] = (char)('0' + vorKomma);
		}

		if((GleitkommaRunden >0 && false==positiv.null) || 
			(GleitkommaRunden==0 && positiv >= real(5L, Anz16bits(1))
		 ))
		{
					// NICHT bei  i=anz_dezimal_mantisse-1  anfangen, denn
					// der erste Wert dezimalMtsMem[0] ist zunchst unbenutzt.
			for(i=anz_dezimal_mantisse ; i>0 ; --i)
				if(dezimalMtsMem[i] < '9')
				{
					++dezimalMtsMem[i];
					break;
				}
				else
					dezimalMtsMem[i]='0';
			if(0==i)
			{
				_10Exp++;	// hier ist normalerweise  _10Exp < 0 !
				dezimalMtsMem[1]='1';
				for(i=2; i<=anz_dezimal_mantisse; i++)
					dezimalMtsMem[i]='0';
			}
		}
	}

	dezimalMtsMem[anz_dezimal_mantisse+1]=0;
	dezimalMtsMem[0]=dezimalMtsMem[1];
	dezimalMtsMem[1]='.';
	strcat(DezimalString, dezimalMtsMem);
	uint8 posE=strlen(DezimalString);
	strcat(DezimalString, "E");
	char exponent_cStr[34];
	_10Exp+=AddToExp10;
	if(0!=ExponentBasis10)
		*ExponentBasis10=_10Exp;
	itoa(_10Exp, (char*)exponent_cStr, 10);
	if(_10Exp > 0)
		strcat(DezimalString, "+");
	strcat(DezimalString, exponent_cStr);
	if(MantisseUndExponentGetrennt)
		DezimalString[posE]=0;

	delete[] dezimalMtsMem;

	if(negativ)
		GleitkommaRunden= - GleitkommaRunden;	// rckgngig von oben.

	return DezimalString;
}

char *real::optimal_decimal_representation(uint8 max_nr_chars, uint8 max_leading_mantisse_zeros/*=3*/)
{
	char *r;
	if(false==null && false==unendlich && (exp2>=0 ? exp2 : -exp2) < ((max_nr_chars*213u)>>6))
	{
		int16 exp10;
		int8 nrd=is_negative() ? max_nr_chars-1 : max_nr_chars; // nrd: Zahl der zu berechnenden dezimalen Mantissestellen.
			// Diese Variable nrd muss sowohl hier als auch unten innerhalb der j-Schleife nach dem to_decimalExp()-Aufruf
			// auf is_negative() ? max_nr_chars-1 : max_nr_chars; initialisiert werden.
		int8 nrd_vor=nrd+1;
		for(uint8 j=0; j<3 && nrd!=nrd_vor; j++)
		{
			if(nrd>nrd_vor)
			{
				r=to_decimalExp(nrd_vor, true, &exp10); // Nur exp10 berechnen
				break;
			}
			r=to_decimalExp(nrd, true, &exp10); // Nur exp10 berechnen
			if(null || unendlich)
				return r;
			if(exp10<-max_leading_mantisse_zeros-1)
				break;
			nrd_vor=nrd;

			nrd=negativ ? max_nr_chars-1 : max_nr_chars; // Diese Variable nrd muss sowohl hier als oben vor Beginn der	
													 // Schleife initialisiert werden.
			if(exp10<0)
				nrd += exp10-1; // exp10-1 Zeichen durch 0.00 .... verbraucht.
			else if(exp10+1<nrd) // && exp10>=0 .
				--nrd; // Ein Zeichen durch Dezimalpunkt verbraucht.
					// Beispiel fr den Fall exp10==max_nr_chars-2 : Es sei max_nr_chars=4, exp10=2, Zahl=683.79426
					// ==> Es wird 684 dargestellt
					// (3 statt 4 Zeichen belegt, weil 683.8 zu viele Zeichen belegen wrde).
		}

		if(!null && exp10>=-max_leading_mantisse_zeros && nrd>exp10)
		{		// Darstellung ohne 10er-Exponent (E)
			uint8 j;
			if(negativ)
				++r;  // - Zeichen
			if(exp10<0)
			{
				for(j=strlen(r) ; (j--)>0 && '0'==r[j] ; )
					r[j]=0;
				r[1]=r[0];
				for(++j ; j>0 ; j--)
					r[j-exp10]=r[j];
				r[0]='0';
				r[1]='.';
				for(j=2; ++exp10<0; )
					r[j++]='0';
			}
			else
			{
			for(j=1 ; j<=exp10 ; j++)
				if(0==(r[j]=r[j+1]))
				{
					while(j<=exp10)
						r[j++]='0';
					r[j]=0;
					break;
				}
				if(j+1<max_nr_chars && 0!=r[j])
				{
					r[j]='.';
					for(j=strlen(r) ; (j--)>0 && '0'==r[j] ; )
						r[j]=0;
					if('.'==r[j])
						r[j]=0;
				}
				else
					r[j]=0;
			}
			if(negativ)
				--r;
			return r;
		}
	}

	// Darstellung mit 10er-Exponent (E)
	uint8 j=max_nr_chars-4;
	if(negativ) --j;
	if(exp2<0) --j;
	if(exp2>34 || exp2<-34) --j;
	if(j<1) j=1;
	while((r=to_decimalExp(j, false)), strlen(r)>(uint8)max_nr_chars)
		if(--j<1)
			break;
	for(j=2; 0!=r[j] && 'E'!=r[j] && 'e'!=r[j] ; j++)
		;
	uint8 k=j;
	while((--j>=4 || !negativ && j>=3) && '0'==r[j])
		;
	while(0!=(r[++j]=r[k++]))
		;
	return r;
}

void real::normalisiere()
		// Die real mu korrekt initialisiert sein mit der
		// Ausnahme, da im Falle null==false und unendlich==false in der Mantisse
		// beliebig viele Nullen vorhanden sein drfen. Insbesondere drfen
		// das hchstwertige Mantissewort und niederwertigere auch Null sein.
		// Auerdem darf vor dem Aufruf von normalisiere() der Exponent 'exp2'
		// auch auerhalb der Grenzen [[min_exp2 , max_exp2]] liegen.
		// Wenn null==true oder unendlich==true ist, so macht normalisiere() nichts.
		// Wenn alle Mantisseworte Null sind, so setzt normalisiere() null=true.
		// Sonst werden die Mantisseworte so verschoben und der Exponent korrekt
		// angepat, da das erste Mantissewort von Null verschieden und mindestens
		// gleich 32768u ist.
		// Wenn am Ende der Exponent kleiner als 'min_exp2' ist, so wird
		// entsprechend der globalen Variablen 'GleitkommaRunden' und der Mantisse
		// entwender die Zahl Null
		// oder die Zahl mit kleinstem zulssigen Betrag und mit gleichem Vorzeichen
		// erzeugt.
		// ACHTUNG: Wenn der Exponent 'exp2' grer als max_exp2 ist, so
		// wird die AUSNAHME OverflowError() geworfen !
{
	if(null || unendlich) return;

	int8 i, j;
	for(i=0; i<anz_mts && mts[i]==0; i++) ;

	if(i==anz_mts)
		null=true;
	else
	{
		exp2-=16*i;
		unsigned short v=32768u;
		for(j=0; mts[i]<v ; v>>=1) ++j;
		exp2-=j;

		int8 anz=i;
		for(i=0 ; i+anz+1<anz_mts ; i++)
			mts[i]=((mts[i+anz]<<j)&65535u) + (mts[i+anz+1]>>(16-j));
		if(i+anz<anz_mts)
			{ mts[i]=((mts[i+anz]<<j)&65535u); i++; }
		for( ; i<anz_mts; i++)
			mts[i]=0;
	}
	check_exp();
}

void real::check_exp()
		// Die real mu korrekt initialisiert sein mit der Ausnahme der
		// Variablen exp2.
		// Falls die real ungleich Null ist und exp2 kleiner ist als
		// min_exp2, so wird je nach der durch 'GleitkommaRunden' bestimmten Rundungsart
		// und je nach den Mantissestellen der Zahl entweder auf Null abgerundet
		// oder aber zur real gerundet, die den kleinstmglichen zulssigen
		// Betrag und das gleiche Vorzeichen wie bisher hat.
{
	if(null || unendlich) return;	// Achtung: Bei null==true ist exp2 nicht gltig !

	if(exp2<min_exp2)
	{
		if(0==GleitkommaRunden)
			{ if(exp2+1 < min_exp2) null=true; }
		else if((GleitkommaRunden<0 && !negativ) || (GleitkommaRunden>0 && negativ))
			null=true;

		if(null==false)
		{		// Die Gleitkommazahl mit dem kleinsten zulssigen Betrag und dem
				// gleichen Vorzeichen wie bisher erzeugen.
			exp2=min_exp2;
			mts[0]=32768u;
			for(uint8 i=1; i<anz_mts; i++)
				mts[i]=0;
		}
	}
	else if(exp2>max_exp2)
	{
#ifdef ENABLE_GLEITKOMMA_EXCEPTIONS
		throw OverflowError();
#else
		if(negativ)
		{
			if(GleitkommaRunden < 0)
				{ unendlich=true; null=false; }
			else
				*this=-MaxPositiveReelleZahl(Anz16bits(anz_mts));
		}
		else
		{
			if(GleitkommaRunden > 0)
				{ unendlich=true; null=false; }
			else
				*this=MaxPositiveReelleZahl(Anz16bits(anz_mts));
		}
	}
#endif
}

void real::UebertragUndRunden(int8 runden,
								bool uebertrag, unsigned short weitere_stellen)
	// Die Bedeutung des Parameters 'runden' entspricht der globalen
	// Variablen 'GleitkommaRunden'; der Parameter 'runden' kann aber
	// unabhngig davon festgelegt werden.
	// Der Parameter 'runden' bezieht sich nur auf den Betrag der real.
	// Wenn z.B.  runden > 0  bergeben wird, so wird der BETRAG ggf. immer
	// aufgerundet.
	// ***** ACHTUNG: DIE GESAMTE FUNKTION  UebertragUndRunden()  LIEST NICHT DAS
	// VORZEICHEN 'negativ' AUS. *****
	// Wenn die real gleich Null oder Unendlich ist (null==true
	// oder unendlich==true), so kehrt diese Funktion sofort zurck. Sonst gilt:
	// Wenn uebertrag==true, so addiert UebertragUndRunden() den Wert
	//       (2 hoch (exp2+1))
	//		===> ZUM BETRAG <==
	// der real, welche korrekt initialisert sein mu mit der Ausnahme,
	// da die Mantisse nicht initialisiert sein mu, also nicht mts[0]>=32768u
	// gelten mu. Wenn die real dann also negativ ist, so wird der Wert
	// von ihr subtrahiert.
	// Wenn aber uebertrag==false, so mu die real vollstndig initialisiert
	// sein; die Mantisse mu also dann auch normalisiert sein.
	//
	// Wenn null==false ist, so luft die Addition von  (2 hoch (exp2+1))  zum
	// Betrag der real darauf hinaus, da alle Bits der binren Mantisse
	// um ein Bit nach rechts verschoben werden. Links kommt als neues hchst-
	// wertigstes Bit eine Eins hinzu. Das Bit, welches rechts herausgeschoben wird,
	// bestimmt zusammen mit dem bergebenen Parameter 'runden' ob eine Aufrundung
	// erfolgt.
	// Wenn allerdings alle bisherigen Mantisseworte den Wert 65535u haben, also alle
	// bisherigen Bits der binren Mantisse Eins sind, und wenn nicht die Option
	// "abrunden" (runden<0) festgelegt wurde, so entstehen die neuen
	// Mantissebits 1,0,0,... , und der Exponent 'exp2' wird insgesamt sogar um 2
	// erhht.
	//
	// Im Wort 'weitere_stellen' soll man das nchste (niederwertige) Mantissewort
	// abspeichern, das sich nicht mehr mit der aktuellen Mantisselnge darstellen
	// lt, aber z.B. bei der Addition zweier Gleitkommazahlen mit unterschiedlichem
	// Exponenten berechnet werden kann. Wenn es nicht existiert (z.B. bei der
	// Addition zweier Gleitkommazahlen mit gleichem Exponenten), so soll Null
	// bergeben werden. Wenn mehr als ein niederwertiges Mantissewort nicht mehr
	// mit der aktuellen Mantisselnge dargestellt werden kann, so soll im Falle
	// von runden>0 (hier: Betrge aufrunden) fr weitere_stellen der Wert Null
	// nur dann bergeben werden, wenn alle dieser Mantisseworte gleich null sind.
	// Im anderen Fall soll das hchstwertigste dieser berzhligen Mantisseworte
	// bergeben werden.
{
	if(null || unendlich) return;

	if(uebertrag)
	{
		uint8 i=anz_mts-1;
		bool u2=false;
		if(runden==0)	// zum nchsten runden
		{
			if(0 != (mts[i] & 1)) u2=true;
		}
		else if(runden > 0)	// ggf. immer aufrunden
		{
			if(0 != weitere_stellen || 0 != (mts[i] & 1))
				u2=true;
		}

		for( ; i>0 ; --i)
			mts[i]=(mts[i]>>1) + ((mts[i-1]<<15)&65535u);
		mts[0]=32768u + (mts[0]>>1);
		++exp2;
		int8 save_r=GleitkommaRunden;
		GleitkommaRunden=runden;
		check_exp();
		GleitkommaRunden=save_r;
		
		if(u2 && false==unendlich)	// Falls unendlich==true, wrde mit
					// naechsthoehererBetrag() ein NaN erzeugt, was hier
					// verhindert werden soll.
			naechsthoehererBetrag();	// Wenn alle Mantissebits gleich eins sind,
				// so wird hier 'exp2' noch einmal erhht und die neue Mantisse
				// mit den Bits 1.0 0 0 0 ... erzeugt.
	}
	else
	{
		if(runden==0)	// zum nchsten runden
		{
			if(weitere_stellen >= 32768u)
				naechsthoehererBetrag();
		}
		else if(runden > 0)
		{					// ggf. immer aufrunden
			if(0 != weitere_stellen)
				naechsthoehererBetrag();
		}
	}
}

real::real(const real& x, Anz16bits anz)
		// Copy-Konstruktor, Redefinition der Genauigkeit mglich.
		// Die Mantissenlnge des konstruierten Objekts betrgt
		// anz.anz mal 16 Bits Mantisselnge.
		// Eventuell Auffllen der Mantisse mit Nullen am Ende oder aber
		// Abschneiden und Runden.
{
	negativ=x.negativ;
	exp2=x.exp2;
	unendlich=x.unendlich;
	null=x.null;

	if(anz.anz < 1) anz.anz=1;
	AnzMantisseMemory=anz_mts=anz.anz;
	mts=new unsigned short[AnzMantisseMemory];

	if(false==x.null && false==x.unendlich)
	{
		if(anz_mts < x.anz_mts)
		{
			uint8 i;
			for(i=0; i<anz_mts; i++)
				mts[i]=x.mts[i];

			if(GleitkommaRunden==0)
					// zur nchtgelegenen Zahl runden
			{
				if(x.mts[i] >= 32768u)
					naechsthoehererBetrag();
			}
			else if((!x.negativ && GleitkommaRunden > 0) ||
				(x.negativ && GleitkommaRunden < 0)
			  )
			{
				for( ; i<x.anz_mts ; i++)
					if(x.mts[i] != 0)
					{
						naechsthoehererBetrag();
						break;
					}
			}
		}
		else
		{
			uint8 i;
			for(i=0; i<x.anz_mts; i++)
				mts[i]=x.mts[i];
			while(i<anz_mts)
				mts[i++]=0;
		}
	}
}

void real::naechsthoehererBetrag()
		// Diese Funktion beachtet die Variable 'GleitkommaRunden'
		// nicht.
		// Wenn null==true ist, wird nichts gemacht.
		// Wenn *this reell und nicht Null ist, wird das Vorzeichen
		// beibehalten und der nchsthhere Betrag eingestellt.
		// Wenn *this reell ist und der Betrag von x der grte
		// darstellbare ist, wird
		//   -- Im Fall enable_real_exceptions==false
		//      das Vorzeichen beibehalten und +Unendlich bzw.
		//      -Unendlich erzeugt,
		//   -- bzw. im Fall enable_real_exceptions==true
		//      die Ausnahme OverflowError() geworfen.
		// Wenn *this schon gleich +Unendlich oder -Unendlich ist,
		// wird NaN erzeugt.
{
	if(null) return;
	if(unendlich)
	{
		null=true;	// null==true UND unendlich==true bedeutet NaN.
		return;
	}
	
	int8 i; // nicht uint8 !!!
	for(i=anz_mts ; --i>=0 ; )
		if(mts[i]<65535u)
			{ ++mts[i]; return; }
		else
			mts[i]=0;

	// Der folgende Programmteil wird nur ausgefhrt, wenn vorher
	// alle Mantissewrter auf 65535u waren.
	mts[0]=32768u;
	for(i=1; i<anz_mts; i++) mts[i]=0;
	if( (++exp2) > max_exp2 )
	{
#if ENABLE_GLEITKOMMA_EXCEPTIONS
		throw OverflowError();
#endif
		unendlich=true;
	}
}

real& real::operator=(const real& x)
{
	if(this != &x)
	{
		if(x.anz_mts > AnzMantisseMemory)
		{
			delete[] mts;
			AnzMantisseMemory=anz_mts=x.anz_mts;
			mts=new unsigned short[AnzMantisseMemory];
		}
		else
			anz_mts=x.anz_mts;

		if(!x.null)		// Effizienz: Bei Null nicht kopieren.
			for(uint8 i=0; i<x.anz_mts; i++) mts[i]=x.mts[i];
		null=x.null;
		negativ=x.negativ;
		unendlich=x.unendlich;
		exp2=x.exp2;
	}
	return *this;
}

real::~real()
{
	delete[] mts;
}

real operator+(const real& x, const real& y)
		// Achtung: Es darf keine Referenz geliefert werden, denn die intern
		// aufgerufene Funktion addiere_betraege() oder subtrahiere_betraege()
		// liefert eine Referenz auf eine statische Variable. Z.B. in einem Ausdruck
		// a+b+c wrde dies zu einem Fehler fhren.
{
	if(x.is_NaN() || y.is_NaN())
		return real::MakeNaN();
	// Die Flle, dass x oder y den Betrag Unendlich hat, werden
	// im Folgenden korrekt behandelt (addiere_betraege() bzw.
	// subtrahiere_betraege().

	int8 max_anz=x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts;
	if(x.null) return real(y, real::Anz16bits(max_anz));
	if(y.null) return real(x, real::Anz16bits(max_anz));

	if(!x.negativ && !y.negativ)
	{
		real& z=addiere_betraege(x, y, GleitkommaRunden);
		z.negativ=false;	// Es kann auch z.null==true sein.
		return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
						// Wert geliefert.
	}
	if(x.negativ && y.negativ)
	{
		real& z=addiere_betraege(x, y, -GleitkommaRunden);
		z.negativ=true;	// Es kann auch z.null==true sein.
		return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
						// Wert geliefert.
	}
	if(x.negativ)	// dann ist y positiv.
	{
		if(vergleiche_betraege(x,y))
		{
			real& z=subtrahiere_betraege(x, y, -GleitkommaRunden);
			z.negativ=!z.negativ;	// Es kann auch z.null==true sein.
				// z.negativ wird von subtrahiere_betraege() als false geliefert,
				// weil der Spezialfall, dass x und y den Betrag Unendlich haben, unten
				// abgehandelt wird.
			return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
							// Wert geliefert.
		}
		else
		{
			real& z=subtrahiere_betraege(y, x, GleitkommaRunden);
				// z.negativ muss so belassen werden, wie es von subtrahiere_betraege()
				// geliefert wird.
				// Bei der Addition (-INF) + (+INF) erhlt man im Falle
				// GleitkommaRunden<0 das Ergebnis -INF.
			return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
						// Wert geliefert.
		}
	}
	else		// x ist positiv, y ist negativ.
	{
		if(vergleiche_betraege(x,y))
		{
			real& z=subtrahiere_betraege(x, y, GleitkommaRunden);
				// z.negativ wird von subtrahiere_betraege() als false geliefert,
				// weil der Spezialfall, dass x und y den Betrag Unendlich haben, unten
				// abgehandelt wird.
			return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
							// Wert geliefert.
		}
		else
		{
			real& z=subtrahiere_betraege(y, x, -GleitkommaRunden);
			z.negativ=!z.negativ;	// Es kann auch z.null==true sein.
				// Bei der Addition (+INF) + (-INF) erhlt man im Falle
				// GleitkommaRunden>0 das Ergebnis +INF.
			return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
							// Wert geliefert.
		}
	}
}

real operator-(const real& x, const real& y)
		// Achtung: Es darf keine Referenz geliefert werden, denn die intern
		// aufgerufene Funktion addiere_betraege() oder subtrahiere_betraege()
		// liefert eine Referenz auf eine statische Variable.
{
	if(x.is_NaN() || y.is_NaN())
		return real::MakeNaN();
	// Die Flle, dass x oder y den Betrag Unendlich hat, werden
	// im Folgenden korrekt behandelt (addiere_betraege() bzw.
	// subtrahiere_betraege().

	int8 max_anz=x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts;
	if(x.null) return -real(y, real::Anz16bits(max_anz));
	if(y.null) return real(x, real::Anz16bits(max_anz));

	if(!x.negativ && y.negativ)
	{
		real& z=addiere_betraege(x, y, GleitkommaRunden);
		z.negativ=false;	// Es kann auch z.null==true sein.
		return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
						// Wert geliefert.
	}
	if(x.negativ && !y.negativ)
	{
		real& z=addiere_betraege(x, y, -GleitkommaRunden);
		z.negativ=true;	// Es kann auch z.null==true sein.
		return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
						// Wert geliefert.
	}
	if(x.negativ)	// dann ist y auch negativ.
	{
		if(vergleiche_betraege(x,y))
		{
			real& z=subtrahiere_betraege(x, y, -GleitkommaRunden);
			z.negativ=!z.negativ;	// Es kann auch z.null==true sein.
				// z.negativ wird von subtrahiere_betraege() als false geliefert,
				// weil der Spezialfall, dass x und y den Betrag Unendlich haben, unten
				// abgehandelt wird.
			return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
							// Wert geliefert.
		}
		else
		{
			real& z=subtrahiere_betraege(y, x, GleitkommaRunden);
				// z.negativ muss so belassen werden, wie es von subtrahiere_betraege()
				// geliefert wird.
				// Bei der Subtraktion (-INF) - (-INF) erhlt man im Falle
				// GleitkommaRunden<0 das Ergebnis -INF.
			return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
							// Wert geliefert.
		}
	}
	else		// x und y sind positiv.
	{
		if(vergleiche_betraege(x,y))
		{
			real& z=subtrahiere_betraege(x, y, GleitkommaRunden);
				// z.negativ wird von subtrahiere_betraege() als false geliefert,
				// weil der Spezialfall, dass x und y den Betrag Unendlich haben, unten
				// abgehandelt wird.
			return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
							// Wert geliefert.
		}
		else
		{
			real& z=subtrahiere_betraege(y, x, -GleitkommaRunden);
			z.negativ=!z.negativ;	// Es kann auch z.null==true sein.
				// Bei der Subtraktion (+INF) - (+INF) erhlt man im Falle
				// GleitkommaRunden>0 das Ergebnis +INF.
			return z;	// Wichtig: kopieren ! Es wird keine Referenz, sondern ein
							// Wert geliefert.
		}
	}
}

real operator-(const real& x)	// einstelliges Minus
		// Achtung: Wert und keine Referenz zurckliefern !
{
	real y=x;
	y.negativ = !y.negativ;
	return y;
}

/*real operator+(const real& x)	// einstelliges Plus
{
	return x;
}*/

real fabs(const real& x)
{
	real help=x;
	help.negativ = false;
	return help;
}

bool operator!=(const real& x, const real& y)
{
	return !(x==y);
}

bool operator==(const real& x, const real& y)
{
	if(x.is_NaN())
		return y.is_NaN();
	if(y.is_NaN())
		return false;	// x.is_NaN()==false

#ifndef ONLY_NAN_NO_INFINITY
	if(x.unendlich)
		return true==y.unendlich && x.negativ==y.negativ;
#endif

	if(x.null)
		return y.null;
	if(y.null) // || y.unendlich)
		return false;	// x.null und x.unendlich sind false.

	if(x.negativ != y.negativ || x.exp2 != y.exp2) return false;
	uint8 i;
	for(i=0; i<x.anz_mts && i<y.anz_mts; i++)
		if(x.mts[i] != y.mts[i]) return false;
	for( ; i<x.anz_mts ; i++)
		if(x.mts[i] != 0) return false;
	for( ; i<y.anz_mts ; i++)
		if(y.mts[i] != 0) return false;
	return true;
}

bool operator>=(const real& x, const real& y)
{
	return !(x < y);
}

bool operator<=(const real& x, const real& y)
{
	return !(x > y);
}

bool operator>(const real& x, const real& y)
{
	if(x.is_NaN() || y.is_NaN())
		return false;

#ifndef ONLY_NAN_NO_INFINITY
	if(x.unendlich)
		return false==x.negativ &&
				(true==y.null || false==y.unendlich || true==y.negativ);
	if(y.unendlich)
		return true==y.negativ;	// Es gilt false==x.unendlich*/
#endif

	if(x.null) return (!y.null) && y.negativ;
	if(y.null) return !x.negativ;		// x.null==false .
	if(!x.negativ)
	{
		if(y.negativ) return true;
		return vergleiche_betraege(x, y);
	}
	else
	{
		if(!y.negativ) return false;
		return vergleiche_betraege(y, x);
	}
}

bool operator<(const real& x, const real& y)
{
	return y>x;
}


bool vergleiche_betraege(const real& x, const real& y)
		// Die Gleitkommazahlen x, y mssen korrekt initialisiert sein.
		// Liefert true, wenn abs(x) > abs(y) gilt.
		// Im Falle abs(x)==abs(y) wird false geliefert.
{
	if(x.is_NaN() || y.is_NaN() || true==y.unendlich)
		return false;
#ifndef ONLY_NAN_NO_INFINITY
	if(x.unendlich)
		return true;	// Es gilt false==y.unendlich
#endif

	if(x.null) return false;
	if(y.null) return !x.null;
	if(x.exp2 > y.exp2) return true;
	if(x.exp2 < y.exp2) return false;
	uint8 i;
	for(i=0; i<x.anz_mts && i<y.anz_mts; i++)
	{
		if(x.mts[i] > y.mts[i]) return true;
		if(x.mts[i] < y.mts[i]) return false;
	}
	for( ; i<x.anz_mts; i++)
		if(x.mts[i] != 0) return true;
	return false;
}

real& addiere_betraege(const real& x, const real& y, int8 runden)
		// Achtung: Es wird eine Referenz auf eine STATISCHE VARIABLE dieser
		// Funktion geliefert. D.h. es mu nach dem Aufruf dieser Funktion
		// noch kopiert werden !
		// ... Die (ber die Referenz) gelieferte real hat immer
		// die Elementvariable negativ==false ; die Variablen x.negativ und
		// y.negativ werden nicht beachtet. Die Variable  null  der gelieferten
		// real wird gesetzt, wenn sich Null ergab.
		//
		// Diese Funktion nimmt unabhngig von der Variablen 'negativ' immer an,
		// da die Gleitkommazahlen x und y positiv sind. Daraus ergibt sich auch, wie
		// der Parameter 'runden' interpretiert wird: Wenn z.B. 'runden' > 0 bergeben
		// wird, so wird der Betrag des Ergebnisses immer aufgerundet.
{
	int8 max_anz=x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts;
	static real *z0=0;
	if(0==z0)
		z0=new real(real::Anz16bits(1));  // wird am Ende des Programms nicht mit delete freigegeben.
	*z0=real(real::Anz16bits(max_anz));	// Die Zuweisung darf
		// nicht direkt bei der Deklaration erfolgen, denn sie mu bei JEDEM Aufruf
		// dieser Funktion erneut erfolgen, was im anderen Fall wegen der statischen
		// Variablen z0 nicht der Fall wre.

	if(x.is_NaN() || y.is_NaN())
		return *z0=real::MakeNaN();
#ifndef ONLY_NAN_NO_INFINITY
	if(x.unendlich)
		return *z0=x; // +Unendlich oder -Unendlich
	if(y.unendlich)
		return *z0=y; // +Unendlich oder -Unendlich*/
#endif

	// --- Achtung ---: folgende Ausdrcke fhren zu falsch initialisierten Zeigern
	// px bzw. py, wenn der rechte Teil ausgefhrt wird (Konstruktion).
	// Denn das konstruierte Objekt ist dann ein temporres Objekt, welches nach
	// beendigung der Programmzeile wieder gelscht wird, so da der Zeiger
	// ungltig ist:
	// FALSCH: real *px=(x.anz_mts==max_anz) ? &x
	//							: &real(x, real::Anz16bits(max_anz));
	// FALSCH: real *py=(y.anz_mts==max_anz) ? &y
	//							: &(real(y, real::Anz16bits(max_anz)));

	real *memX=0, *memY=0;
	const real *px=(x.anz_mts==max_anz) ? &x
								: (memX=new real(x, real::Anz16bits(max_anz)));
	const real *py=(y.anz_mts==max_anz) ? &y
								: (memY=new real(y, real::Anz16bits(max_anz)));
	
	if(px->exp2 < py->exp2)	// Denn unten wird x0.exp2 >= y0.exp2 vorausgesetzt.
	{
		const real *help=px;
		px=py;
		py=help;
	}

	const real& x0=*px;
	const real& y0=*py;
	
		// Nun haben x0 und y0 die gleiche Mantissenlnge, nmlich max_anz.
	
	if(x0.null)
	{
		if(memX) delete memX;
		if(memY) delete memY;
		return (*z0=y0);	// max_anz Mantisseworte
	}
	if(y0.null)
	{
		if(memX) delete memX;
		if(memY) delete memY;
		return (*z0=x0);	// max_anz Mantisseworte
	}

	int16 anz, i, d;

	anz=max_anz - (x0.exp2 - y0.exp2) / 16;	// Es ist anz <= max_anz, denn
															// x0.exp2 >= y0.exp2 (siehe oben).
		// Es kann auch  anz <= 0  sein. In diesem Fall mu der Wert z0=x0 zurck-
		// gegeben werden, wenn keine Aufrundung stattfindet. Ansonsten mu
		// der gegenber x0 entsprechend erhhte Wert geliefert werden.
	d=(x0.exp2 - y0.exp2) & 15;

	unsigned short weitere_stelle=0;
		// Im folgenden wird zwar ermittelt, ob ein Aufrunden stattfinden soll
		// aufgrund der Mantissestellen, die nicht mehr dargestellt werden knnen.
		// Im Falle eines Aufrundens wird dieser Wert aber nicht in der folgenden
		// Addition der Mantissen von x0 und y0 bercksichtigt. Denn wenn sich
		// durch diese Addition ein bertrag ber das bisher hchstwertigste hinaus
		// Mantissebit ergibt, so mu bei 'runden'==0 nur das niederwertigste Bit,
		// das noch in der Addition bercksichtigt wurde, untersucht werden. Ist
		// es gleich eins, so wird aufgerundet. Dies wird durch den unteren
		// Aufruf  z0->UebertragUndRunden()  erledigt.
		// Durch diese Technik, den im folgenden berechneten Wert fr die Variable
		// 'weitere_stelle' nicht sofort bei der Addition der Mantissen fr eine
		// eventuelle Aufrundung zu verwenden, wird ein sonst mgliches
		// DOPPELTES RUNDEN VERMIEDEN.

	if(runden==0)		// den Betrag zum nchsten Wert hin runden.
	{
		if(0==d)
			{ if(anz>=0 && anz<max_anz)  weitere_stelle=y0.mts[anz]; }
					// Im Fall anz==max_anz darf y0.mts[anz] nicht gebildet werden.
		else
			{ if(anz>=1)  weitere_stelle=((y0.mts[anz-1]<<(16-d))&65535u); }
					// Hier ist immer anz-1 < max_anz .
	}
	else if(runden>0)
	{		// Betrag aufrunden
		if(d>0 && anz>=1  &&  0 != ((y0.mts[anz-1]<<(16-d))&65535u))
			weitere_stelle=1;
		else
		{
			if((i=anz) < 0) i=0;
			for( ; i<y0.anz_mts ; i++)
				if(0 != y0.mts[i])
					{ weitere_stelle=1; break; }
		}
	}
			// sonst: Betrag abrunden ==> nichts tun
		
	// Wenn hier schon anz<=0 gilt, so wird die folgende (erste) Schleife nicht
	// ausgefhrt und statt dessen in der zweiten Schleife die Mantisse von x0 nach
	// z0 kopiert.
	long uebertrag=0, zwischen;

	i=max_anz-1;

	if(anz > 0)
	{
		for( ; --anz > 0 ; --i)
			if(  (zwischen = ((long)x0.mts[i]) + uebertrag +
									((long)(y0.mts[anz]>>d)) +
									((long)( (y0.mts[anz-1]<<(16-d))&65535u ))
				  ) >= 65536L
			  )
				{ z0->mts[i] = (unsigned short)(zwischen-65536L);  uebertrag=1; }
			else
				{ z0->mts[i] = (unsigned short)zwischen;  uebertrag=0; }
		// Nun ist anz==0 .
		if(  (zwischen = ((long)x0.mts[i]) + uebertrag +
						     ((long)(y0.mts[anz]>>d))
		     ) >= 65536L
	     )
			{ z0->mts[i] = (unsigned short)(zwischen-65536L);  uebertrag=1; }
		else
			{ z0->mts[i] = (unsigned short)zwischen;  uebertrag=0; }
		--i;
	}

	for( ; i>=0 ; --i)
		if((zwischen = ((long)x0.mts[i]) + uebertrag) >= 65536L)
			{ z0->mts[i] = (unsigned short)(zwischen-65536L);  uebertrag=1; }
		else
			{ z0->mts[i] = (unsigned short)zwischen;  uebertrag=0; }

	z0->exp2=x0.exp2;
		// exp2 wird durch den Aufruf UebertragUndRunden() noch erhht,
		// wenn uebertrag!=0 ist.
	z0->null=false;		// Dies mu vor UebertragUndRunden() erfolgen.
				// Wenn beide Summanden Null sind, so erfolgte schon ein return.
	z0->UebertragUndRunden(runden, 0 != uebertrag, weitere_stelle);

	z0->normalisiere();
		// Testet auf einen zu kleinen Exponenten und rundet eventuell. Wirft bei
		// zu groem Exponenten die Ausnahme OverflowError().

	if(memX) delete memX;
	if(memY) delete memY;
	
	return *z0;	// Achtung: Referenz auf statische Variable !
}

real& subtrahiere_betraege(const real& x, const real& y,
												 int8 runden)
		// Achtung: Es wird eine Referenz auf eine STATISCHE VARIABLE dieser
		// Funktion geliefert. D.h. es mu nach dem Aufruf dieser Funktion
		// noch kopiert werden !
		// ACHTUNG: Es werden nur die Betrge von x und y beachtet.
		// --- DER BETRAG VON x DARF NICHT KLEINER SEIN ALS DER VON y !!!
		// ... Die (ber die Referenz) zurckgelieferte real hat immer
		// die Elementvariable negativ==false ; die Variablen x.negativ und
		// y.negativ werden nicht beachtet. Die Variable  null  der gelieferten
		// real wird gesetzt, wenn sich Null ergab.
		//
		// Diese Funktion nimmt unabhngig von der Variablen 'negativ' immer an,
		// da die Gleitkommazahlen x und y positiv sind. Auerdem wird vorausgesetzt,
		// da der Betrag von x nicht kleiner ist als der Betrag von y.
		// Daraus ergibt sich auch, wie der Parameter 'runden' interpretiert wird:
		// Wenn z.B. 'runden' > 0 bergeben wird, so wird der Betrag des Ergebnisses
		// ggf. immer aufgerundet. Die zurckgegebene Zahl ist nicht negativ.
		// EINZIGE AUSNAHME: WENN SOWOHL x ALS AUCH y GLEICH PLUS UNENDLICH SIND
		// UND WENN runden<0 BERGEBEN WIRD, DANN WIRD DIE ZAHL MINUS UNENDLICH
		// ZURCKGEGEBEN.
{
	int8 max_anz=x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts;
	static real *z0=0;
	if(0==z0)
		z0=new real(real::Anz16bits(1));  // wird am Ende des Programms nicht mit delete freigegeben.
	*z0=real(real::Anz16bits(max_anz));	// Die Zuweisung darf
		// nicht direkt bei der Deklaration erfolgen, denn sie mu bei JEDEM Aufruf
		// dieser Funktion erneut erfolgen, was im anderen Fall wegen der statischen
		// Variablen z0 nicht der Fall wre.

	// --- Achtung ---: folgende Ausdrcke fhren zu falsch initialisierten Zeigern
	// px bzw. py, wenn der rechte Teil ausgefhrt wird (Konstruktion).
	// Denn das konstruierte Objekt ist dann ein temporres Objekt, welches nach
	// beendigung der Programmzeile wieder gelscht wird, so da der Zeiger
	// ungltig ist:
	// FALSCH: real *px=(x.anz_mts==max_anz) ? &x
	//							: &real(x, real::Anz16bits(max_anz));
	// FALSCH: real *py=(y.anz_mts==max_anz) ? &y
	//							: &(real(y, real::Anz16bits(max_anz)));

	if(x.is_NaN() || y.is_NaN())
		return *z0=real::MakeNaN();
#ifndef ONLY_NAN_NO_INFINITY
	if(x.unendlich)
	{
		if(y.unendlich)
		{
			if(runden == 0)
				return *z0;	// z0 == 0
			else if(runden < 0)
				return *z0=real::MakeMinusInfinity();
			return *z0=real::MakePlusInfinity();
		}
		return *z0=real::MakePlusInfinity();
	}
#endif

	real *memX=0, *memY=0;
	const real *px=(x.anz_mts==max_anz) ? &x
								: (memX=new real(x, real::Anz16bits(max_anz)));
	const real *py=(y.anz_mts==max_anz) ? &y
								: (memY=new real(y, real::Anz16bits(max_anz)));
	
	if(px->exp2 < py->exp2)	// Denn unten wird x0.exp2 >= y0.exp2 vorausgesetzt.
	{
		const real *help=px;
		px=py;
		py=help;
	}

	const real& x0=*px;
	const real& y0=*py;
	
		// Nun haben x0 und y0 die gleiche Mantissenlnge, nmlich max_anz.
	
	if(x0.null)
	{
		if(memX) delete memX;
		if(memY) delete memY;
		return *z0;	// z0==0 (Default nach der Konstruktion) . Wenn x0==0 ist,
			// so darf subtrahiere_betraege() nur aufgerufen werden, wenn auch y0==0 ist.
			// z0 hat max_anz Mantisseworte.
	}
	if(y0.null)
	{
		if(memX) delete memX;
		if(memY) delete memY;
		return (*z0=x0);	// max_anz Mantisseworte.
	}

	int16 anz, i, d;

	anz=max_anz - (x0.exp2 - y0.exp2) / 16;	// Es ist anz <= max_anz, denn
															// x0.exp2 >= y0.exp2 (siehe oben).
		// Es kann auch  anz <= 0  sein. In diesem Fall mu der Wert z0=x0 zurck-
		// gegeben werden, wenn keine Aufrundung stattfindet. Ansonsten mu
		// der gegenber x0 entsprechend erhhte Wert geliefert werden.
	d=(x0.exp2 - y0.exp2) & 15;

	long uebertrag=0;	// Im Gegensatz zu addiere() kann man bei subtrahiere() ein
		// eventuelles Abrunden gleich bei der Ausfhrung der Subtraktion mit
		// bercksichtigen, denn, falls eine Abrundung notwendig ist, so haben
		// x0 und y0 NICHT den gleichen Wert 'exp2' (ihre Mantissen wurden gleich
		// lang gemacht). Dann hat das Ergebnis aber den gleichen Wert 'exp2' wie
		// x0. Somit ist keine zweite Rundung aufgrund einer verschobenen Mantisse
		// des Ergebnisses notwendig.

	if(runden==0)		// den Betrag zum nchsten Wert hin runden.
	{
		if(0==d)
			{ if(anz>=0 && anz<max_anz && y0.mts[anz] >= 32768u)  uebertrag=1; }
					// Im Fall anz==max_anz darf y0.mts[anz] nicht gebildet werden.
		else
			{ if(anz>=1 && ((y0.mts[anz-1]<<(16-d))&65535u) >= 32768u)  uebertrag=1; }
					// Hier ist immer anz-1 < max_anz .
	}
	else if(runden<0)
	{		// Betrag abrunden (y0 wird subtrahiert !).
		if(d>0 && anz>=1  &&  0 != ((y0.mts[anz-1]<<(16-d))&65535u))
			uebertrag=1;
		else
		{
			if((i=anz) < 0) i=0;
			for( ; i<y0.anz_mts ; i++)
				if(0 != y0.mts[i])
					{ uebertrag=1; break; }
		}
	}
			// sonst: Betrag aufrunden ==> nichts tun
		
	// Wenn hier schon anz<=0 gilt, so wird die folgende (erste) Schleife nicht
	// ausgefhrt und statt dessen in der zweiten Schleife die Mantisse von x0 nach
	// z0 kopiert.
	long zwischen;
	
	i=max_anz-1;

	if(anz > 0)
	{
		for(i=max_anz-1 ; --anz > 0 ; --i)
			if(  (zwischen = ((long)x0.mts[i]) - uebertrag
							     - ((long)(y0.mts[anz]>>d))
							     - ((long)( (y0.mts[anz-1]<<(16-d))&65535u ))
			     ) < 0
		     )
				{ z0->mts[i] = (unsigned short)(zwischen+65536L);  uebertrag=1; }
			else
				{ z0->mts[i] = (unsigned short)zwischen;  uebertrag=0; }
		// Nun ist anz==0 .
		if(  (zwischen = ((long)x0.mts[i]) - uebertrag
						     - ((long)(y0.mts[anz]>>d))
		     ) < 0
	     )
			{ z0->mts[i] = (unsigned short)(zwischen+65536L);  uebertrag=1; }
		else
			{ z0->mts[i] = (unsigned short)zwischen;  uebertrag=0; }
		--i;
	}

	for( ; i>=0 ; --i)
		if((zwischen = ((long)x0.mts[i]) - uebertrag) < 0)
			{ z0->mts[i] = (unsigned short)(zwischen+65536L);  uebertrag=1; }
		else
			{ z0->mts[i] = (unsigned short)zwischen;  uebertrag=0; }

				// Wenn diese Funktion korrekt benutzt wird, so ist an dieser
				// Stelle uebertrag==0, denn der Betrag von x0 darf nicht kleiner
				// als der Betrag von y0 sein.

	z0->exp2=x0.exp2;
		// exp2 wird durch den Aufruf UebertragUndRunden() noch erhht,
		// wenn uebertrag!=0 ist.
	z0->negativ=false;
	z0->null=false;		// Dies mu vor normalisiere() erfolgen.
				// Wenn beide Summanden Null sind, so erfolgte schon ein return.
	z0->normalisiere();
		// Normalisiert die Mantisse, falls durch die Subtraktion fhrende Nullen
		// entstanden sind, wenn also mts[0] < 32768u gilt.
		// Testet auf einen zu kleinen Exponenten und rundet eventuell. Wirft bei
		// zu groem Exponenten die Ausnahme OverfolwError().

	if(memX) delete memX;
	if(memY) delete memY;
	
	return *z0;	// Achtung: Referenz auf statische Variable !
}

real operator*(const real& x, const real& y)
{
	return multipliziere(x, y);	// kopiert von der Referenz auf die statische
		// Variable in multipliziere(), die von multipliziere() geliefert wird,
		// in eine temporre Variable hinein, die zurckgeliefert wird.
};

real& multipliziere(const real& x, const real& y)
		// multipliziere() bercksichtigt auch die Vorzeichen der vollstndig
		// initialisierten Zahlen x und y.
		// Achtung: liefert Referenz auf statische Variable.
{
	static real *z0=0;
	if(0==z0)
		z0=new real(real::Anz16bits(1));  // wird am Ende des Programms nicht mit delete freigegeben.
										// wenig Speicher verbrauchen: 1 Mantissewort.
	int16 nx=x.AnzMtsBenutzt(), ny=y.AnzMtsBenutzt();
	int16 max_anz=x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts; // nicht nx, ny, denn durch
		// x.anz_mts > x.nx wird angezeigt, dass die gewnschte Genauigkeit durch x.anz_mts gegeben ist.

	*z0=real(real::Anz16bits(max_anz));	// Die Zuweisung darf
		// nicht direkt bei der Deklaration erfolgen, denn sie mu bei JEDEM Aufruf
		// dieser Funktion erneut erfolgen, was im anderen Fall wegen der statischen
		// Variablen *z0 nicht der Fall wre.

	if(x.is_NaN() || y.is_NaN())
		return *z0=real::MakeNaN();
#ifndef ONLY_NAN_NO_INFINITY
	if(x.unendlich)
	{
		if(y.null)
		{
			if(GleitkommaRunden > 0)
				*z0=real::MakePlusInfinity();
			else if(GleitkommaRunden < 0)
				*z0=real::MakeMinusInfinity();
			// sonst: *z0=0 ( 0 * Unendlich == 0 bei
			// 				  GleitkommaRunden==0 )
			return *z0;
		}
		bool neg_erg=x.negativ;
		if(y.negativ)
			neg_erg=!neg_erg;
		*z0=x;
		z0->negativ=neg_erg;
		return *z0;
	}
	if(y.unendlich)
	{
		if(x.null)
		{
			if(GleitkommaRunden > 0)
				*z0=real::MakePlusInfinity();
			else if(GleitkommaRunden < 0)
				*z0=real::MakeMinusInfinity();
			// sonst: z0=0 ( 0 * Unendlich == 0 bei
			// 				  GleitkommaRunden==0 )
			return *z0;
		}
		bool neg_erg=y.negativ;
		if(x.negativ)
			neg_erg=!neg_erg;
		*z0=y;
		z0->negativ=neg_erg;
		return *z0;
	}
#endif

	const real *px = &x;
	const real *py = &y;

	if(nx < ny)
	{		// nur zur Effizienz beim Multiplizieren
		const real *help=px;
		px=py;
		py=help;
		int16 h=nx;
		nx=ny;
		ny=h;
	}

	const real& x0=*px;
	const real& y0=*py;
		// x0 und y0 knnen verschiedene Mantisselngen haben !!!
	
	if(x0.null || y0.null)
		{ z0->null=true; return *z0; }	// z0 hat max_anz Mantisseworte.

	/*if(1==ny)
	{
		if(32768u==y0.mts[0])
		{	// Abfrage, ob y0 gleich 2^n oder -2^n ist. ACHTUNG: Nach dem Ausschluss von x0.null || y0.null !!!
			*z0 = real(x0, real::Anz16bits(max_anz));
			z0->exp2 += y0.exp2;
			if(y0.negativ)
				z0->negativ = !z0->negativ;
			return *z0;
		}
		if(1==nx && 32768u==x0.mts[0])
		{	// Abfrage, ob x0 gleich 2^n oder -2^n ist. ACHTUNG: Nach dem Ausschluss von x0.null || y0.null !!!
			*z0 = real(y0, real::Anz16bits(max_anz));
			z0->exp2 += x0.exp2;
			if(x0.negativ)
				z0->negativ = !z0->negativ;
			return *z0;
		}
	}*/

	if((!x0.negativ && !y0.negativ) || (x0.negativ && y0.negativ))
		z0->negativ=false;
	else
		z0->negativ=true;

	uint16 i,j;
	
	int16 anz_lm=nx+ny;
	unsigned long *langeMantisse=new unsigned long[anz_lm+2];
		// Es mssen zwei Worte mehr als anz_lm reserviert werden, weil unten beim Runden eventuell auf diese
		// Worte zugegriffen wird (sie werden im Folgenden mit Nullen gefllt).
	for(i=0; i<anz_lm+2; i++)
		langeMantisse[i]=0;

	for(i=0; i<ny; i++)
	{
		register unsigned long m=(unsigned long)y0.mts[i], uebertrag=0;
		long k=nx+i; j=nx;
		if(k>max_anz+1)
			{ j-=k-(max_anz+1); k=max_anz+1; }
		while((j--)>0) // Es ist immer k>=j, daher braucht nicht auf k>=0 getestet zu werden.
		{
			register unsigned long h=uebertrag + m*((unsigned long)x0.mts[j]) +
												((unsigned long)langeMantisse[k]);
			langeMantisse[k--] = (h & 65535U);
			uebertrag=(h >> 16);
		}
		do
		{
			register unsigned long h=uebertrag + ((unsigned long)langeMantisse[k]);
			langeMantisse[k--] = (h & 65535u);
			uebertrag=(h >> 16);
		} while(uebertrag!=0 && k>=0);
			// Der Fall uebertrag!=0 UND k<0 tritt nie ein. Die Abfrage k>=0
			// ist nur vorsichtshalber vorhanden.
	}

	z0->null=false;		// VOR einem eventuellen Aufruf von 'naechsthoehererBetrag()' !

	for(i=max_anz+2; true; i++)
	{	// Im Fall i==nx+ny wird ggf. nur noch der Code fr Runden, nicht aber zur Berechnung
		// von langeMantisse[] ausgefhrt.
		unsigned long maxerr=(nx+ny-i) + 1 + ((2*(nx+ny-i))>>real::bits_per_digit);
		if(i>max_anz+2)
			maxerr = 1 + (maxerr >> real::bits_per_digit);
		if(i>=nx+ny)
			maxerr=0;

		if(0 != (langeMantisse[0] & 32768u))
		{
			z0->exp2=x0.exp2 + y0.exp2 + 1;	// Test auf Unter- bzw. berlauf erfolgt unten.
			for(j=(max_anz<anz_lm) ? max_anz : anz_lm ; (j--)>0 ; )
				z0->mts[j]=(unsigned short)langeMantisse[j];
			for(j=anz_lm ; j<max_anz; j++)
				z0->mts[j]=0;

			if(max_anz < anz_lm) // der Code fr den Fall 0==GleitkommaRunden setzt
					// max_anz<anz_lm voraus (Zugriff auf langeMantisse[max_anz] !).
			{
				if(0==GleitkommaRunden)
				{
					if(langeMantisse[max_anz] >= 32768u && langeMantisse[max_anz]+maxerr < 3LU*32768LU)
						{ z0->naechsthoehererBetrag(); break; }
					else if(langeMantisse[max_anz]+maxerr < 32768u)
						break;
				}
				else if((GleitkommaRunden>0 && !z0->negativ) || (GleitkommaRunden<0 && z0->negativ))
				{		// Betrag aufrunden
					if(langeMantisse[max_anz]+maxerr<65536LU)
					{
						for(j=max_anz; j<anz_lm ; j++)
							if(0 != langeMantisse[j])
								{ z0->naechsthoehererBetrag(); break; }
						if(j<anz_lm || 0==maxerr)
							break;
					}
				}
				else if(langeMantisse[max_anz]+maxerr<65536LU)
					break;	// Betrag abrunden ==> nichts tun
			}
		}
		else
		{
			z0->exp2=x0.exp2 + y0.exp2;		// Test auf Unter- bzw. berlauf erfolgt unten.
			for(j=(max_anz<anz_lm) ? max_anz : anz_lm ; (j--)>0 ; )
				z0->mts[j]=(unsigned short)( ((langeMantisse[j]<<1)&65535u) + (langeMantisse[j+1]>>15) );
			for(j=anz_lm ; j<max_anz; j++)
				z0->mts[j]=0;

			if(max_anz < anz_lm) // der Code fr den Fall 0==GleitkommaRunden setzt
					// max_anz<anz_lm voraus (Zugriff auf langeMantisse[max_anz] !).
			{
				if(0==GleitkommaRunden)
				{
					if((langeMantisse[max_anz] & 32767) >= 16384 && (langeMantisse[max_anz] & 32767)+maxerr < 3u*16384u)
						{ z0->naechsthoehererBetrag(); break; }
					else if((langeMantisse[max_anz] & 32767)+maxerr < 16384)
						break;
				}
				else if((GleitkommaRunden>0 && !z0->negativ) || (GleitkommaRunden<0 && z0->negativ))
				{		// Betrag aufrunden
					if((langeMantisse[max_anz] & 32767)+maxerr < 32768u)
					{
						if(0 != (langeMantisse[max_anz] & 32767)) // das hchstwertigste Bit nicht testen !
							{ z0->naechsthoehererBetrag(); break; }
						else
						{
							for(j=max_anz + 1 ; j<anz_lm ; ++j)
								if(0 != langeMantisse[j])
									{ z0->naechsthoehererBetrag(); break; }
							if(j<anz_lm || 0==maxerr)
								break;
						}
					}
				}
				else if((langeMantisse[max_anz] & 32767)+maxerr < 32768u)
					break;	// Betrag abrunden ==> nichts tun
			}
		}

		if(i>=nx+ny)
			break;
		for(j=i-nx ; j<ny ; j++)
		{
			register unsigned long h=((unsigned long)y0.mts[j]) * ((unsigned long)x0.mts[i-1-j]) + langeMantisse[i];
			langeMantisse[i] = (h & 65535U);
			if( (h = ((unsigned long)langeMantisse[i-1]) + ((h >> 16) & 65535U)) < 65536LU )
				langeMantisse[i-1]=h;
			else
			{
				langeMantisse[i-1]=(h & 65535U);
				for(long k=i-1 ; --k>=0 ; )
				{
					if(65535U!=langeMantisse[k])
						{ ++langeMantisse[k]; break; }
					langeMantisse[k]=0;
				}
			}
		}
	}

	z0->normalisiere();	// null=false ist schon oben erfolgt.
		// Testet auf einen zu kleinen Exponenten und rundet eventuell. Wirft bei
		// zu groem Exponenten die Ausnahme OverflowError().

	delete[] langeMantisse;

	return *z0;
}

real operator/(const real& x, const real& y)
{
#ifdef DIVISION_DURCH_NEWTON_ITERATION
	return dividiere_newton_iterationen(x, y, 0);		// kopiert von der Referenz auf die statische
		// Variable in multipliziere(), die von multipliziere() geliefert wird,
		// in eine temporre Variable hinein, die zurckgeliefert wird.
#else
	return dividiere(x, y, 0);		// kopiert von der Referenz auf die statische
		// Variable in multipliziere(), die von multipliziere() geliefert wird,
		// in eine temporre Variable hinein, die zurckgeliefert wird.
#endif
};

#ifndef DIVISION_DURCH_NEWTON_ITERATION
unsigned long approx_high_ulongword_of_ulong_mult_ulong(unsigned long x, unsigned long y)
{
	return (x>>16LU)*(y>>16LU) + (((x>>16LU)*(y&0xffffLU))>>16LU) + (((y>>16LU)*(x&0xffffLU))>>16LU);
}

unsigned long approx_inverse_of_fixpoint_number(unsigned long y)
{	// Liefert eine Approximation fr (2.0 hoch 31) * (  1.0 / (y / (2.0 hoch 31))  )
	// Voraussetzung: y muss zwischen 0x800000000 und 0xffffffff liegen (entspricht zwischen 1.0 und kleiner als 2.0).
	unsigned long z=0xffffffffLU / (y>>16LU);
	return (approx_high_ulongword_of_ulong_mult_ulong(z<<14LU, (0x80000000LU - approx_high_ulongword_of_ulong_mult_ulong(z<<14LU,y))<<1LU)) << 1LU;
}

real& dividiere(const real& x, const real& y, real **rest)
		// Achtung: liefert Referenz auf statische Variable.
		// Default fr 'rest' ist der Nullzeiger.
		// dividiere() bercksichtigt auch die Vorzeichen der vollstndig
		// initialisierten Zahlen x und y.
		// --- Achtung ---: dividiere() liefert eine Referenz auf statische Variable.
		//
		// Wenn rest!=0 ist, so mu es auf einen Zeiger auf eine real
		// zeigen. Dann schreibt dividiere in diesen Zeiger *rest die Adresse
		// einer STATISCHEN real dieser Funktion (dividiere()), in
		// die der Rest der Division abgespeichert wird. Der Rest hat das
		// gleiche Vorzeichen wie 'x', sofern der (genherte) zurckgegebene
		// Quotient nicht vom Betrag aufgerundet wird.
		// Der Rest wird im Falle einer Aufrundung des Betrags des Quotienten
		// entsprechend abgendert, so da u.a. auch der Rest sein Vorzeichen gegenber
		// dem Fall ndert, da keine Aufrundung stattfindet.
		// Das bedeutet, mit den folgenden Programmzeilen kann man eine
		// Division durchfhren und diese durch anschlieende Multiplikation
		// (mit ausreichender Mantisselnge) und Addition des Rests EXAKT umkehren:
		//
		//    bool f(real x, real y)
		//    {
		//       real *rest_ptr, rest;
		//       real z=dividiere(x, y, &rest_ptr);
		//       rest=*rest_ptr;   // kopieren von der statischen Variablen in
		//                         // dividiere().
		//      	real z2(z,
		//             real::Anz16bits(z.get_anz_mts() + y.get_anz_mts())
		//                        );
		//                 // z2 entsteht aus z durch Erweitern der Mantisse mit Null-
		//                 // worten. Dies ist notwendig, damit bei bei der folgenden
		//                 // Multiplikation z2*y keine Rundungen erfolgen mssen.
		//       real x2=z2*y + rest;
		//       return x==x2;   // Liefert im Normalfall 'true'.
		//    }

		// Durch eine Umkehrung der Division durch eine Multplikation des (gerundeten)
		// und auf eine ausreichende Mantisselnge erweiterten Quotienten z2 mit y
		// und anschlieender Addition des Restes erhlt man EXAKT
		// den Wert x zurck, vorausgesetzt
		// es trat kein berlauf oder Unterlauf durch einen zu groen oder einen
		// zu kleinen Exponenten des Ergebnisses auf.
		// --- Zur Beachtung ---: Der Quotient hat als Anzahl der Mantissestellen
		// immer das Maximum der Zahl der Mantissestellen der bergebenen Werte x
		// und y. Falls 'rest' ungleich Null bergeben wird, so hat die statische
		// real **rest danach soviele Mantissestellen wie y.
{
	static real *z0=0;
	if(0==z0)
		z0=new real(real::Anz16bits(1));  // wird am Ende des Programms nicht mit delete freigegeben.
	static real *StaticRest=0;
	if(0==StaticRest)
		StaticRest=new real(real::Anz16bits(1));  // wird am Ende des Programms nicht mit delete freigegeben.
	*StaticRest=real(real::Anz16bits(y.anz_mts));
		// ==> Unten wird teilweise vorausgesetzt, dass *StaticRest mit Null initialisiert wurde <==
		// Obwohl der RestSpeicher 'y.anz_mts' + 1 Worte aufnehmen kann (abgesehen
		// von 3 zustzlichen Nullen am Ende), sind nach Durchfhrung des letzten
		// Divisionsschritts (Schleife for(i ...) {} ) davon maximal die
		// 'y.anz_mts' hchstwertigsten von Null verschieden, so da diese Anzahl
		// an Mantissestellen fr 'StaticRest' ausreicht. Denn bereits ein voriger
		// Divisionsschritt hat die letzte Stelle der Mantisse von x0 "heruntergeholt",
		// so da im letzten Divisionsschritt immer eine Null "heruntergeholt" wird.
		// Voraussetzung ist, da (wie in dieser Funktion) der Quotient immer soviele
		// Mantissestellen hat wie das Maximum der Anzahl der Mantissestellen von
		// x0 und y0.
	if(rest)
		*rest=StaticRest;

	uint16 max_anz=x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts;
	*z0=real(real::Anz16bits(max_anz));	// Die Zuweisung darf
		// nicht direkt bei der Deklaration erfolgen, denn sie mu bei JEDEM Aufruf
		// dieser Funktion erneut erfolgen, was im anderen Fall wegen der statischen
		// Variablen *z0 nicht der Fall wre.

	if(x.is_NaN() || y.is_NaN())
	{
		if(rest) *StaticRest=real::MakeNaN();
		return *z0=real::MakeNaN();
	}

#ifndef ONLY_NAN_NO_INFINITY
	if(x.unendlich)
	{
		if(y.null)
		{
			// Auch im Falle GleitkommaRunden==0 wird Unendlich erzeugt,
			// denn nur +Unendlich und -Unendlich sind mglich.
			if(GleitkommaRunden >= 0)
				*z0=real::MakePlusInfinity();
			else
				*z0=real::MakeMinusInfinity();
			return *z0;
		}

		bool neg_erg=x.negativ;
		if(y.negativ)
			neg_erg=!neg_erg;

		if(y.unendlich)
		{
			if(0==GleitkommaRunden)
				*z0=real(1L, real::Anz16bits(max_anz));
				// Unendlich / Unendlich == 1 bei GleitkommaRunden==0
			else if((false==neg_erg && GleitkommaRunden > 0)
					|| (true==neg_erg && GleitkommaRunden < 0))
				*z0=real::MakePlusInfinity();
			// sonst: *z0=0 ( Unendlich / Unendlich = 0
			//               bei GleitkommaRunden < 0 )
			
			z0->negativ=neg_erg;
			if(rest && !x.negativ && GleitkommaRunden>0) *StaticRest=real::MakePlusInfinity();
			if(rest && x.negativ && GleitkommaRunden<0) *StaticRest=real::MakeMinusInfinity();
			return *z0;
		}

		*z0=real::MakePlusInfinity();
		z0->negativ=neg_erg;
		if(rest)
		{
			if((!x.negativ && GleitkommaRunden>0) || (x.negativ && GleitkommaRunden<0))
				*StaticRest=y;
			StaticRest->negativ=x.negativ;
		}
		return *z0;
	}

	if(y.unendlich)	// x.unendlich==false
		return *z0=real(real::Anz16bits(max_anz)); // Null*/
#endif

	const real& x0=x;
	const real& y0=y;

		// x0 und y0 knnen verschiedene Mantissenlnge haben !!!!

	if(y0.mts[0]<32768u)	// Vorliegen einer nicht normalisierten Zahl.
	{
#if ENABLE_GLEITKOMMA_EXCEPTIONS
		throw OverflowError();
#endif
		return *z0=real::MakeNaN();
	}

	if(y0.null)
	{
#if ENABLE_GLEITKOMMA_EXCEPTIONS
		throw real::DivisionByZeroError();
#endif
#ifdef ONLY_NAN_NO_INFINITY
		return *z0=real::MakeNaN();
#else
		// Auch im Falle GleitkommaRunden==0 wird Unendlich erzeugt,
		// denn nur +Unendlich und -Unendlich sind mglich.
		if(GleitkommaRunden >= 0)
			return *z0=real::MakePlusInfinity();
		return *z0=real::MakeMinusInfinity();
#endif
	}

	if(x0.null)
		{ z0->null=true; return *z0; }	// *z0 hat max_anz Mantisseworte.

	/*if(1==y0.AnzMtsBenutzt() && 32768u==y0.mts[0])
	{	// Abfrage, ob y0 gleich 2^n oder -2^n ist. ACHTUNG: Nach dem Ausschluss von x0.null || y0.null !!!
		*z0 = real(x, real::Anz16bits(max_anz));
		z0->exp2 -= y0.exp2;
		if(y0.negativ)
			z0->negativ = !z0->negativ;
		// *StaticRest wurde oben auf Null gesetzt und braucht hier nicht verndert zu werden,
		// denn eine Division durch 2^n kann immer ohne Rundungsfehler erfolgen.
		return *z0;
	}*/

	if((!x0.negativ && !y0.negativ) || (x0.negativ && y0.negativ))
		z0->negativ=false;
	else
		z0->negativ=true;

	// Weil die Gleitkommazahlen x0 und y0 normalisiert sind, so gilt
	// 32768u <= x0.mts[0] < 65536LU  und  32768u <= y0.mts[0] < 65536LU .
	// Der (genherte) Quotient (und der Rest) mssen auch normalisiert
	// vorliegen. Daher werden die Bits der Mantisse von x0 ggf. so verschoben,
	// da das erste Wort des Quotienten bereits zwischen 32768u und 65535u liegt.
	// Denn andernfalls, wenn es z.B. kleiner als 32768u wre, so mte
	// ein Wort fr den Quotienten mehr berechnet werden und dann die Bits
	// des Ergebnisses so verschoben werden, da die Normalisierung hergestellt
	// wird.
	// Wenn die Mantisse von x0 kleiner ist als diejenige von y0, so liegt das
   // erste Quotientenwort bereits zwischen 32768u und einschlielich 65535u.
   // Andernfalls werden die Bits der
	// Mantisse von x0 vor der Ausfhrung des Divisionsalgorithmus (in Gedanken)
	// um eine Position nach rechts verschoben. Dies bewirkt eine Division
	// durch Zwei, die spter in Exponenten 'exp2' des Ergebnisses bercksichtigt
	// wird. Tatschlich jedoch wird die Verschiebung nur jedes Mal vorgenommen,
	// wenn ein Mantissewort von x0 ausgelesen werden soll.
	//
	// Im folgenden wird die Variable 'x0_div2' auf 'true' gesetzt, wenn x0
	// (vom Betrage her) ursprnglich eine grere oder gleich groe Mantisse
	// hatte als y0.

	int i,j;
	
	bool x0_div2=true, fertig=false;
	for(i=0; i<x.anz_mts && i<y.anz_mts; i++)
	{
		if(x.mts[i] > y.mts[i]) {                fertig=true; break; }
		if(x.mts[i] < y.mts[i]) { x0_div2=false; fertig=true; break; }
	}
	if(!fertig)
		for( ; i<y.anz_mts; i++)
			if(y.mts[i] != 0)
				{ x0_div2=false; }
	// Falls nun x0_div2==true ermittelt wurde, so mu statt jeden Zugriffs
	// auf x0.mts[n], der bei x0_div==false stattfindet, der Zugriff
	//         (x0.mts[n]>>1)                       falls    n==0   
	// bzw.    (x0.mts[n]>>1) + ((x0.mts[n-1]<<15)&65535u)
	//                                              falls    0 < n < x0.anz_mts
	// bzw.    ((x0.mts[n-1]<<15)&65535u)            falls    n==x0.anz_mts
	// erfolgen.
	// Insbesondere entsteht fr n==x0.anz_mts ein neues Mantissewort, das
	// im Fall x0_div2==false nicht vorhanden ist. Auch im Fall x0_div2==false
	// gilt die folgende Aussage:
	//
	// Wenn wie in dieser Divisionsroutine der Quotient mit sovielen
	// Mantissewrtern berechnet wird, wie
	// das Maximum von x0.anz_mts und y0.anz_mts ist, so wurden bereits VOR dem
	// letzten Divisionsschritt alle Mantissewrter von x0 "heruntergeholt", so
	// da zur Abspeicherung des Rests (Parameter 'rest' von dividiere()) immer
	// y0.anz_mts Mantissewrter ausreichen.

	int y0_anz=y0.anz_mts;
	int RestAnz=y0_anz + 1;
	long *RestSpeicher=new long[RestAnz + 2];
		// WICHTIG: 2 Worte mehr reservieren und diese mit Nullen fllen (unten).
		// Sie werden eventuell bei der Berechnung von Dx (unten) bentigt.

		// RestSpeicher[0] hat beim Test, wie oft die Mantisse von y0 in RestSpeicher[]
		// enthalten ist, eine um den Faktor 65536 grere Wertigkeit als y0.mts[0].

	int x0_index;

	RestSpeicher[0]=0;
	if(false==x0_div2)
		for(x0_index=0; x0_index<x0.anz_mts && x0_index<y0_anz; x0_index++)
			RestSpeicher[x0_index+1]=x0.mts[x0_index];
	else
	{
		RestSpeicher[1] = (x0.mts[0] >> 1);		// x0.mts[0] mu laut Voraussetzung
				// ber die Klasse real immer existieren.
		for(x0_index=1; x0_index<x0.anz_mts && x0_index<y0_anz; x0_index++)
			RestSpeicher[x0_index+1] =
						(x0.mts[x0_index]>>1) + ((x0.mts[x0_index-1]<<15) & 65535u);
		if(x0_index==x0.anz_mts && x0_index<y0_anz)
		{
			RestSpeicher[x0_index+1]=((x0.mts[x0_index-1]<<15) & 65535u);
			x0_index++;
		}
	}
	// ACHTUNG: x0_index darf nun nicht mehr verndert werden. Diese Variable
	// wird erst in der unteren Divisionshauptschleife weiterverwendet und gendert.
	for(i=x0_index ; i+1<RestAnz + 2 ; i++)	// wichtig: +3 , auch mit Nullen fllen.
		RestSpeicher[i+1]=0;

	z0->exp2=x0.exp2 - y0.exp2;
	if(x0_div2) ++z0->exp2;	// Falls die Mantisse des ursprnglichen x0 grer oder
				// gleich ist als die von y0, so mute sie durch 2 dividiert werden.
				// Dies wird kompensiert durch Inkrementierung des Exponenten 'exp2'
				// des Quotienten.
		// Unten wird z0->exp2 noch dekrementiert, falls die (GEDACHTE, EVENTUELL
		// DURCH ZWEI DIVIDIERTE) Mantisse von x0 NUN kleiner ist als die Mantisse
		// von y0. Dies tritt in dieser Programmversion IMMER auf.
		// Der Test auf berlauf (Overfolw-Error) oder Unterlauf erfolgt unten.

	unsigned long invDy=((unsigned long)y0.mts[0])<<16L;
	if(y0_anz>=2) invDy+=y0.mts[1];
	invDy=approx_inverse_of_fixpoint_number(invDy);

	bool AbsMtsX0KleinerAbsMtsY0=false;		// Ist erst nach dem ersten
			// Schleifendurchlauf korrekt.
	bool erstes;

	for(i=0, erstes=true ; i<max_anz ; erstes=false)
					// Bei einer fhrenden Null im Ergebnis 'i' nicht erhhen !
	{
			// Nun wird getestet, wie oft die Zahl in der Mantisse von y0
			// in RestSpeicher enthalten ist. Nur wenn die Variable 'enthalten'
			// nach der Rechnung den Wert Null enthlt, so ist dieser Wert
			// mit Sicherheit richtig. Andernfalls kann es vorkommen, da
			// die Mantisse von y0 tatschlich maximal einmal mehr als
			// 'enthalten' mal in RestSpeicher enthalten ist. Dies mu dann
			// nach der Subtraktion des Wertes
			//			'enthalten' * (Mantisse von y0)
			// von RestSpeicher durch einen Grenvergleich festgestellt werden,
			// um eventuell noch einmal den Wert 'Mantisse von y0' abzuziehen.
		uint16 enthalten;
		
		unsigned long Dx=(((unsigned long)RestSpeicher[0])<<16L) + RestSpeicher[1];
		Dx = approx_high_ulongword_of_ulong_mult_ulong(Dx, invDy);
		enthalten = (Dx<16) ? 0 : (Dx-16L)>>14L;

			// Nun wird im Falle 'enthalten==0' getestet, ob die Mantisse von y0
			// tatschlich kein Mal in RestSpeicher enthalten ist oder ob sich
			// der Wert 0 nur durch Abrunden ergab. In letzterem Fall wird enthalten
			// auf 1 gesetzt.
		if(0==enthalten)
		{
			if(0 != RestSpeicher[0])
				enthalten=1;
			else
			{
				for(j=0; j<y0_anz; j++)
					if((long)y0.mts[j] < RestSpeicher[j + 1])
						{ enthalten=1; break; }
					else if((long)y0.mts[j] > RestSpeicher[j + 1])
						break;	// enthalten=0
					else if(j==y0_anz-1)
						{ enthalten=1; break; }		// Gleichheit.
			}
		}

		if(erstes && 0==enthalten)		// In dieser Programmversion, die
					// mit normalisierten Mantissen arbeitet, knnte man auch nur
					// die Abfrage  if(erstes)  machen.
		{
			--z0->exp2;
			AbsMtsX0KleinerAbsMtsY0=true;		// Dies wird spter bei der Abspeicherung
					// des Rests in *StaticRest bentigt. Diese Variable wurde noch
					// aus einer alten Programmversion beibehalten, die mit nicht
					// notwendigerweise normalisierten Mantissen arbeitete. In dieser
					// Programmversion wird  AbsMtsX0KleinerAbsMtsY0=true;  immer im
					// ersten Divisionsschritt ausgefhrt. Unter 'X0' ist hierbei
					// die durch 2 geteilte Mantisse von x0 zu verstehen, wenn die
					// ursprngliche Mantisse von x0 grer ist als die von y0.
			for(j=0; j<y0_anz; j++)
				RestSpeicher[j]=RestSpeicher[j+1];

			if(x0_index > x0.anz_mts || (x0_index==x0.anz_mts && false==x0_div2))
				RestSpeicher[y0_anz]=0;
			else if(x0_index < x0.anz_mts)
			{
				if(false==x0_div2)
					RestSpeicher[y0_anz]=x0.mts[x0_index++];
				else
				{
						// Hier ist aufgrund der Initialisierungen vor dieser i-Schleife
						// sicher schon x0_index > 0 . Daher kann x0_index-1 immer
						// gebildet werden.
					RestSpeicher[y0_anz] = (x0.mts[x0_index]>>1) +
													((x0.mts[x0_index - 1]<<15) & 65535u);
					x0_index++;
				}
			}
			else	// x0_index==x0.anz_mts && true==x0_div2
			{
				RestSpeicher[y0_anz]=((x0.mts[x0.anz_mts-1]<<15) & 65535u);
				x0_index++;
			}

			// i bleibt hier auf Null, und z0->mts[i] wird noch nicht beschrieben !
		}
		else
		{
			long save=RestSpeicher[y0_anz], uebertrag=0;
			bool abziehen=false;

				// Der folgende Programmteil zur Zuweisung an RestSpeicher[y0_anz]
				// ist identisch wie der entsprechende in obigem
				// 'if(erstes && enthalten==0)' - Block.
			if(x0_index > x0.anz_mts || (x0_index==x0.anz_mts && false==x0_div2))
				RestSpeicher[y0_anz]=0;
			else if(x0_index < x0.anz_mts)
			{
				if(false==x0_div2)
					RestSpeicher[y0_anz]=x0.mts[x0_index++];
				else
				{
						// Hier ist aufgrund der Initialisierungen vor dieser i-Schleife
						// sicher schon x0_index > 0 . Daher kann x0_index-1 immer
						// gebildet werden.
					RestSpeicher[y0_anz] = (x0.mts[x0_index]>>1) +
													((x0.mts[x0_index - 1]<<15) & 65535u);
					x0_index++;
				}
			}
			else	// x0_index==x0.anz_mts && true==x0_div2
			{
				RestSpeicher[y0_anz]=((x0.mts[x0.anz_mts-1]<<15) & 65535u);
				x0_index++;
			}

               // Wichtig: 'save' mu den Wert VOR dem gerade durchlaufenden
               // if-else-Block haben !
			for(j=y0_anz; --j>=0 ; )
			{
				register unsigned long m=((unsigned long)enthalten) *
												 ((unsigned long)y0.mts[j]);
				register long zwischen=save - ((long)(m & 65535LU)) - uebertrag;
				uebertrag=(m >> 16);
				while(zwischen < 0)
					{ uebertrag++; zwischen+=65536L; }
				save=RestSpeicher[j];
				RestSpeicher[j]=zwischen;
			}
               // An dieser Stelle ist  save-uebertrag  gleich dem hchstwertigsten
               // Wort von RestSpeicher, welches theoretisch an der Position
               // RestSpeicher[-1] stehen mte. Wenn der Wert 'enthalten' richtig
               // war, so ist dieses Wort gleich Null. Es kann aber auch der Fall
               // eintreten, da 'enthalten' um eins zu klein war. Dann ist 
               //   save-uebertrag > 0  , und die Mantisse von y0 mu noch einmal
               // von RestSpeicher abgezogen werden.
			if(save > uebertrag)
				abziehen=true;
			else
			{
				// Weil eventuell 'enthalten' um eins zu klein ist, so mu in
				// einem solchen Fall nun noch einmal die Mantisse von y0 von
				// RestSpeicher[] (ohne das neu hinzugekommene RestSpeicher[y0_anz])
				// abgezogen werden.
				// Nun wird getestet, ob dieses Restspeicher[] ohne RestSpeicher[y0_anz]
				// noch grer oder gleich wie y0.mts[] ist.
				for(j=0; j<y0_anz; j++)
					if((long)y0.mts[j] < RestSpeicher[j])
						{ abziehen=true; break; }
					else if((long)y0.mts[j] > RestSpeicher[j])
						break;	// enthalten=0
					else if(j==y0_anz-1)
						{ abziehen=true; break; }		// Gleichheit.
			}

			if(abziehen)
			{
	            ++enthalten;
				long uebertrag=0, zwischen;
				for(j=y0_anz ; --j>=0 ; )
					if((zwischen = RestSpeicher[j] - ((long)y0.mts[j]) - uebertrag) < 0)
						{ RestSpeicher[j] = zwischen+65536L;  uebertrag=1; }
					else
						{ RestSpeicher[j] = zwischen;  uebertrag=0; }
			}

			z0->mts[i] = (unsigned short)enthalten;
			i++;
		}
	}

	long uebertrag=0;
	
	if(GleitkommaRunden==0)		// zum nchsten Wert runden
	{
			// Es wird aufgerundet, wenn der doppelte Rest grer oder gleich wie
			// die Mantisse von y0 ist.
		if(RestSpeicher[0] >= 32768u)
			uebertrag=1;
		else
		{
			for(j=0; j<y0_anz; j++)
			{
				unsigned long h=(((unsigned long)RestSpeicher[j]) & 32767u)*2u +
										((unsigned long)RestSpeicher[j+1]) / 32768u;
				if(h < (unsigned long)y0.mts[j])
					break;
				else if(h > (unsigned long)y0.mts[j] ||
						 (h==(unsigned long)y0.mts[j] && j==y0_anz-1))
					{ uebertrag=1; break; }
			}
		}
	}
	else if((GleitkommaRunden>0 && !z0->negativ) || (GleitkommaRunden<0 && z0->negativ))
							// aufrunden
	{
		for(j=0; j<RestAnz; j++)
			if(0!=RestSpeicher[j])
				{ uebertrag=1; break; }
	}

	if(rest)
	{
				// *StaticRest ist durch eine obige Zuweisung mit Null
				// (und mit y0.anz_mts Mantissestellen) initialisiert worden.
		StaticRest->exp2=x0.exp2 - 16*z0->anz_mts;
		if(false==AbsMtsX0KleinerAbsMtsY0)
			StaticRest->exp2 += 16;	// Dieser Fall tritt in dieser Programmversion mit
						// der Verwendung von normalisierten Mantissen nie ein.
		if(true==x0_div2)
			++StaticRest->exp2;	// Denn: In diesem Fall wurde die durch zwei geteilte
					// Mantisse von x0 verwendet.

		// Eine eventuell notwendige Normalisierung der Mantisse des Rests erfolgt unten.
	}
	
	if(uebertrag)
	{		// Den Quotienten *z0 aufrunden.
		long zwischen;
		for(i=max_anz ; --i>=0 ; )
			if((zwischen=1L+z0->mts[i]) >= 65536L)
				z0->mts[i]=0;
			else
				{ z0->mts[i]=(unsigned short)zwischen; break; }
		if(i < 0)
		{
			++z0->exp2;
			z0->mts[0]=32768u;
			for(i=1; i<max_anz; i++)
				z0->mts[i]=0;
			// Rechts wird durch die Normalisierung nur eine Null herausgeschoben,
			// so da die Mantissenlnge von z0 weiterhin ausreicht.
		}

		if(rest)
		{
				// Nach einer Aufrundung des Ergebnisses wird der Betrag des Rests
				// entsprechend verkleinert. Dadurch wird der (ursprnglich als
				// positiv angenommene) Restwert negativ.

			long u2=0;
				// Hier gilt RestSpeicher[y0.anz_mts]==0 .
			for(i=y0_anz ; --i>=0 ; )
			{
				long zwischen;
				if((zwischen=((long)y0.mts[i]) - RestSpeicher[i] - u2) >= 0)
				{
					StaticRest->mts[i]=(unsigned short)zwischen;
					u2=0;
				}
				else
				{
					StaticRest->mts[i]=(unsigned short)(zwischen+65536L);
					u2=1;
				}
			}
			StaticRest->negativ= ! x0.negativ;
		}
	}
	else	// kein Aufrunden des Quotienten
	{
		if(rest)
		{
			for(i=y0_anz ; --i>=0 ; )
				StaticRest->mts[i]=(unsigned short)RestSpeicher[i];
			StaticRest->negativ=x0.negativ;
		}
	}

	if(rest)
	{
		StaticRest->null=false;	// ist VOR normalisiere() notwendig und kann durch
			// diese Funktion eventuell wieder rckgngig gemacht werden.
		StaticRest->normalisiere();
		*rest=StaticRest;
	}

	z0->null=false;		// VOR einem eventuellen z0->check_exp() !

	z0->normalisiere();
		// Testet auf einen zu kleinen Exponenten und rundet ggf. entweder auf Null
		// oder auf den kleinsten darstellbaren Betrag. Wirft bei
		// zu groem Exponenten die Ausnahme OverflowError().

	delete[] RestSpeicher;
	return *z0;
}
#endif

real fmod(const real& x, const real& y,
						  real *ganz)
		// ..... Default fr ganz ist der Null-Zeiger.
		// ----- ACHTUNG -----: DIESE FUNKTION ARBEITET NUR KORREKT, WENN GILT:
		//     abs(x) <= abs(y) * (2 hoch (max_anz*16))
		// Dabei ist 'max_anz' das Maximum der Anzahlen der Mantissewrter
		// der Zahlen x und y (jedes Mantissewort hat 16 Bits).
		// Auerdem mu  y != 0  gelten.
		// -------------------
		// Wenn diese Bedingungen erfllt sind, so werden eine ganze Zahl  n  und
		// eine real  f  berechnet, so da
		//      x == n*y + f    und     0 <= abs(f) < y
		// gelten und  f  das gleiche Vorzeichen hat wie x.
		// Die Zahl  f  wird von fmod() zurckgeliefert. Wenn  ganz != 0  bergeben
		// wird, so wird nach  *ganz  die ganze Zahl  n  geschrieben.
		// ===> Es spielt fr die Ergebnisse von fmod() keine Rolle, welchen
		//      Wert die Variable 'GleitkommaRunden' vor dem Aufruf von fmod()
		//      hatte. Natrlich hat 'GleitkommaRunden' nach dem Aufruf von
		//      fmod() den gleichen Wert wie vor dem Aufruf.
		//
		// ===> Wenn obige Bedingungen an x und y erfllt sind, SO GILT DIE GLEICHUNG
		//        x == n*y + f  EXAKT, WENN MAN VORAUSSETZT, DASS DIE
		//      RECHNUNG  n*y + f  MIT MINDESTENS  max_anz + y.anz_mts
		//      Mantissewrtern zu je 16 Bits durchgefhrt
		// wird, wobei  max_anz  das Maximum der Zahl der Mantissewrter
		// von x und y sei und weiter  y.anz_mts  die Zahl der Mantissewrter von y.
		// Das heit, vor der Durchfhrung der Rechnung  n*y + f  mte man die
		// von fmod() gelieferten Werte  n  und  f  auf  max_anz + y.anz_mts  Mantisse-
		// wrter erweitern (mit ResizeMantisseAnz16Bits()), denn:
		//
		// Sowohl der Rckgabewert f als auch die eventuell nach  *ganz  geschriebene
		// Zahl  n  besitzen soviele Mantissewrter wie das Maximum der Anzahl
		// der Mantissewrter der Zahlen x und y.
{
	if(x.unendlich || y.unendlich)
		return real::MakeNaN();
	int16 max_anz=x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts;

	int8 AltGleitkommaRunden=GleitkommaRunden;
		// Die folgende Einstellung von 'GleitkommaRunden' hngt vom Vorzeichen
		// des Quotienten  x/y  ab und bewirkt, da immer ein Abschneiden (KEIN
		// Aufrunden) Restes (der zunchst nicht ganzzahlig sein mu) erfolgt.
	GleitkommaRunden=-1;
	if((x.negativ && !y.negativ) || (!x.negativ && y.negativ))
		GleitkommaRunden=1;
	real Ganz=to_integer(x/y);
	GleitkommaRunden=AltGleitkommaRunden;

	if(ganz) *ganz=Ganz;
	return (x-y*Ganz.ResizeMantisseAnz16Bits(max_anz + y.anz_mts)).ResizeMantisseAnz16Bits(max_anz);
}

#ifdef DIVISION_DURCH_NEWTON_ITERATION
real& dividiere_newton_iterationen(const real& x, const real& y, real* rest)
{	// ACHTUNG: Liefert eine Referenz auf eine statische Variable !
	// Obwohl diese Funktion den Quotienten durch eine Newton-Iteration berechnet, ist das Rundungsverhalten
	// exakt gleich wie bei der Funktion dividiere(). Wenn z.B. der Quotient x/y durch die Anzahl der
	// Mantissestellen, welche dem Maximum der Anzahlen von x und y entspricht, exakt dargestellt werden kann,
	// ist auch im Fall GleitkommaRunden=-1 und x>0, y>0 sichergestellt, dass dieser exakte Quotient und
	// nicht etwa ein sich durch Subtraktion einer kleinen Zahl ergebender Wert zurckgeliefert wird.
	// Dies ist insbesondere bei der Funktion fmod() wichtig, welche dividiere() aufruft, und innerhalb von
	// dividiere() wird dividiere_newton_iterationen() bei hoher Anzahl von Mantissestellen aufgerufen.

	static real *z0=0;
	if(0==z0)
		z0=new real(real::Anz16bits(1));  // wird am Ende des Programms nicht mit delete freigegeben.
	static real *StaticRest=0;
	if(0==StaticRest)
		StaticRest=new real(real::Anz16bits(1));  // wird am Ende des Programms nicht mit delete freigegeben.
	*StaticRest=real(real::Anz16bits(y.anz_mts));
		// ==> Unten wird teilweise vorausgesetzt, dass *StaticRest mit Null initialisiert wurde <==
		// Obwohl der RestSpeicher 'y.anz_mts' + 1 Worte aufnehmen kann (abgesehen
		// von 3 zustzlichen Nullen am Ende), sind nach Durchfhrung des letzten
		// Divisionsschritts (Schleife for(i ...) {} ) davon maximal die
		// 'y.anz_mts' hchstwertigsten von Null verschieden, so da diese Anzahl
		// an Mantissestellen fr 'StaticRest' ausreicht. Denn bereits ein voriger
		// Divisionsschritt hat die letzte Stelle der Mantisse von x0 "heruntergeholt",
		// so da im letzten Divisionsschritt immer eine Null "heruntergeholt" wird.
		// Voraussetzung ist, da (wie in dieser Funktion) der Quotient immer soviele
		// Mantissestellen hat wie das Maximum der Anzahl der Mantissestellen von
		// x0 und y0.
	if(rest)
		*rest=*StaticRest;
	if(rest) *rest=real(real::Anz16bits(y.anz_mts));

	if(x.is_NaN() || y.is_NaN())
	{
		if(rest) *rest=real::MakeNaN();
		return *z0=real::MakeNaN();
	}

#ifndef ONLY_NAN_NO_INFINITY
	if(x.unendlich)
	{
		if(y.null)
		{
			// Auch im Falle GleitkommaRunden==0 wird Unendlich erzeugt,
			// denn nur +Unendlich und -Unendlich sind mglich.
			if(GleitkommaRunden >= 0)
				*z0=real::MakePlusInfinity();
			else
				*z0=real::MakeMinusInfinity();
		}

		bool neg_erg=x.negativ;
		if(y.negativ)
			neg_erg=!neg_erg;

		if(y.unendlich)
		{
			if(0==GleitkommaRunden)
				*z0=real(1L, real::Anz16bits((x.anz_mts > y.anz_mts) ? x.anz_mts : y.anz_mts));
				// Unendlich / Unendlich == 1 bei GleitkommaRunden==0
			else if((false==neg_erg && GleitkommaRunden > 0)
					|| (true==neg_erg && GleitkommaRunden < 0))
				*z0=real::MakePlusInfinity();
			// sonst: *z0=0 ( Unendlich / Unendlich = 0
			//               bei GleitkommaRunden < 0 )
			
			z0->negativ=neg_erg;
			if(rest && !x.negativ && GleitkommaRunden>0) *rest=real::MakePlusInfinity();
			if(rest && x.negativ && GleitkommaRunden<0) *rest=real::MakeMinusInfinity();
			return *z0;
		}

		*z0=real::MakePlusInfinity();
		z0->negativ=neg_erg;
		if(rest)
		{
			if((!x.negativ && GleitkommaRunden>0) || (x.negativ && GleitkommaRunden<0))
				*rest=y;
			rest->negativ=x.negativ;
		}
		return *z0;
	}

	if(y.unendlich)	// x.unendlich==false
		return *z0=real(real::Anz16bits((x.anz_mts > y.anz_mts) ? x.anz_mts : y.anz_mts)); // Null*/
#endif
		// x0 und y0 knnen verschiedene Mantissenlnge haben !!!!

	if(y.mts[0]<32768u)	// Vorliegen einer nicht normalisierten Zahl.
	{
#ifdef ENABLE_GLEITKOMMA_EXCEPTIONS
		throw real::DivisionByZeroError();
#endif
		return *z0=real::MakeNaN();
	}

	if(y.null)
	{
#ifdef ENABLE_GLEITKOMMA_EXCEPTIONS
		throw real::DivisionByZeroError();
#endif
		// Auch im Falle GleitkommaRunden==0 wird Unendlich erzeugt,
		// denn nur +Unendlich und -Unendlich sind mglich.
#ifndef ONLY_NAN_NO_INFINITY
		if(GleitkommaRunden >= 0)
			return *z0=real::MakePlusInfinity();
		return *z0=real::MakeMinusInfinity();
#else
		return *z0=real::MakeNaN();
#endif
	}

	if(x.null)
		{ z0->null=true; return *z0; }	// *z0 hat max_anz Mantisseworte.

	/*if(1==y.AnzMtsBenutzt() && 32768u==y.mts[0])
	{	// Abfrage, ob y gleich 2^n oder -2^n ist. ACHTUNG: Nach dem Ausschluss von x.null || y.null !!!
		*z0 = real(x, real::Anz16bits(x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts));
		z0->exp2 -= y.exp2;
		if(y.negativ)
			z0->negativ = !z0->negativ;
		return *z0;
	}*/

	int8 runden_vorher=GleitkommaRunden;
	GleitkommaRunden=0;	// wird unten rckgngig gemacht.

	real y0=y;

	int16 soll_genau=16*(x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts) + 6 + 4;
	*z0=real(real::AnzBits(soll_genau));

	z0->ResizeMantisseAnz16Bits(1);		// Der weitere Speicher bleibt zugewiesen; deshalb
		// mu bei einer folgenden schrittweisen Erhhung der Genauigkeit der
		// Mantisse von y nicht immer der bisherige Speicher freigegeben und neuer
		// zugewiesen werden.
		// ACHTUNG: ResizeMantisseAnz16Bits() darf nicht zwischen den unten stehenden
		// Anweisungen, welche die Mantisse und den Exponenten von y verndern,
		// stehen, weil ResizeMantisseAnz16Bits() immer eine korrekt initialisierte Zahl
		// voraussetzt; insbesondere fhrt ResizeMantisseAnz16Bits beim Verkleinern
		// der Mantissenlnge ggf. Rundungen durch.

	if(y0.mts[0]==32768u)
	{
		z0->mts[0]=32768u;
		z0->exp2=-y0.exp2;
	}
	else
	{
		z0->mts[0]=(1LU<<31LU) / ((unsigned long)y0.mts[0]);
		z0->exp2=-y0.exp2-1;
	}
	
	z0->null=z0->negativ=false; // z0->negativ wird noch korrigiert

	// z0 ist nun (bis auf den zu addierenden Exponenten add_res_exp2)
	// mit einer Genauigkeit von  2 hoch -40  gleich 1/x.
	// Bei dieser Aussage wurden sicherheitshalber etwa 8 hintere Bits
	// nicht bercksichtigt.

	// Wegen der obigen Initialisierung gilt nun y.negativ==false und
	//    y.exp2 == 0 ,
	// also  1 <= y < 2 .

	int16 genau=12, diff, mask;		// mu mindestens gleich 2 sein, damit die folgende
		// Schleife keine Endlosschleife ist.

	real y0_kurz(real::Anz16bits(y0.anz_mts));
	bool first=true;

	while(genau < soll_genau)		// Newton-Iteration
	{
		//printf("genau=%d   soll_genau=%d\n", genau, soll_genau);
		genau=2*genau-1;		// *2 : Verdoppelung der Genauigkeit bei jedem
			// Interationsschritt, wenn dieser aus den Ausgangsdaten (die die
			// Genauigkeit  2 hoch -'genau'  haben) durch Rechenschritte mit
			// beliebig hoher Genauigkeit ausgefhrt wrden.
			// Durch die Verkleinerung von 'genau' um 1 werden Rundungsfehler in der Iteration bercksichtigt.

		if(first)
		{
			int16 g=soll_genau;
			while(g>genau)
				g=(g/2)+1;
			genau=g;
			for(mask=1 ; g<soll_genau; g=2*g-1)
				mask <<= 1;
			diff=g-soll_genau;
			first=false;
		}
		else if(0!=(diff & (mask>>=1)))
			--genau;

		if(genau > soll_genau) genau=soll_genau;		// Die +1 ist auch
			// fr das Abbruchkriterium dieser Schleife wichtig (es erfolgt noch
			//   --genau) !

		int16 anzb=genau + 5;	// Dies bewirkt, da die Rechnung mit
			// mindestens  'genau'+5  Mantissebits durchgefhrt wird, WOBEI
			// MIT  'genau'  DER WERT DIESER VARIABLEN GEMEINT IST, DER AN
			// DIESER PROGRAMMSTELLE, ALSO VOR DEM '--genau' ANGENOMMEN WIRD.

		y0_kurz=y0; y0_kurz.negativ=false;
		y0_kurz.ResizeMantisse(real::AnzBits(anzb));

		*z0 = *z0 * (real(2L, real::Anz16bits(1)) - (y0_kurz*(*z0)));
	};

	*z0 = *z0 * x;

	if((!x.negativ && !y.negativ) || (x.negativ && y.negativ))
		z0->negativ=false;
	else
		z0->negativ=true;

	GleitkommaRunden=(z0->negativ ? 1 : -1);
	z0->RoundUpDown(soll_genau);
		// Nun ist noch eine Genauigkeit von  2 hoch -(16*x0.anz_mts+2)
		// sichergestellt.
	z0->ResizeMantisseAnz16Bits(x.anz_mts > y.anz_mts ? x.anz_mts : y.anz_mts);
	int16 nanz=z0->anz_mts + y.anz_mts, zanz=z0->anz_mts;

	real r;
	if(x.anz_mts >= y.anz_mts)
		r = (fabs(x) - real(fabs(y), real::Anz16bits(nanz))*fabs(*z0))
				.ResizeMantisseAnz16Bits(1+y.anz_mts);
	else
		r = (fabs(x) - real(fabs(y), real::Anz16bits(2*y.anz_mts))*fabs(*z0))
				.ResizeMantisseAnz16Bits(1+y.anz_mts);

	while(r >= real(fabs(y)).mult2hochX(1+z0->exp2-zanz*real::bits_per_digit))
	{
		r -= real(fabs(y)).mult2hochX(1+z0->exp2-zanz*real::bits_per_digit);
		z0->naechsthoehererBetrag(); // nach der nderung von r, weil zuvor noch auf das alte z0->exp2 zugegriffen
			// werden muss.
	}
	while(r < real(0L)) // nicht r.negativ, denn bei r.null==true ist r.negativ nicht definiert.
	{
		r += real(fabs(y)).mult2hochX(1+z0->exp2-zanz*real::bits_per_digit);
		z0->naechstkleinererBetrag(); // nach der nderung von r, weil zuvor noch auf das alte z0->exp2 zugegriffen
			// werden muss.
	}

	long uebertrag=0;
	
	GleitkommaRunden=runden_vorher;

	if(0==GleitkommaRunden)		// zum nchsten Wert runden
	{
		if(real(r).mult2hochX(1) >= real(fabs(y)).mult2hochX(1+z0->exp2-zanz*real::bits_per_digit))
			uebertrag=1;
	}
	else if(!r.null && ((GleitkommaRunden>0 && !z0->negativ) || (GleitkommaRunden<0 && z0->negativ)))
		uebertrag=1;

	if(uebertrag)
	{		// Den Quotienten z0 aufrunden.
		r -= real(fabs(y)).mult2hochX(1+z0->exp2-zanz*real::bits_per_digit);
		z0->naechsthoehererBetrag(); // nach der nderung von r, weil zuvor noch auf das alte z0->exp2 zugegriffen
			// werden muss.
	}

	if(x.negativ) // In der letzten if-Abfrage kann r auch negativ geworden sein. Daher darf man
		r.negativ=!r.negativ; // nicht einfach r.negativ=x.negativ setzen.

	r.ResizeMantisseAnz16Bits(y.anz_mts);

	if(rest)
		*rest=r;

	return *z0;
}
#endif

real& real::operator*=(const real& y)
{
	real& res=multipliziere(*this, y);
	*this=res;
	return *this;
}

real& real::operator/=(const real& y)
{
#ifdef DIVISION_DURCH_NEWTON_ITERATION
	real& res=dividiere_newton_iterationen(*this, y);
#else
	real& res=dividiere(*this, y);
#endif
	*this=res;
	return *this;
}

real& real::operator+=(const real& y)
{
	return ( *this = *this+y );
}

real& real::operator-=(const real& y)
{
	return ( *this = *this-y );
}

real sqrt(const real& x)
		// Wenn x positiv ist, so wird ein positives y geliefert, fr das
		// (nherungsweise) y*y = x gilt.
		// Wenn x negativ ist, so wird ein negatives y geliefert, fr das
		// (nherungsweise) y*y = -x gilt.
		// Wenn man den bergebenen Wert x als exakt betrachtet, so wird
		// durch sqrt(x) die Quadratwurzel von x mit einer Genauigkeit
		// angenhert, die nicht schlechter ist als
		//   2 hoch -(16*x.anz_mts - 2) .
		// Die Variable GleitkommaRunden wird beachtet.
{
	if(x.null || x.unendlich)
		return x;

	int8 runden_vorher=GleitkommaRunden;
	GleitkommaRunden=0;	// wird unten rckgngig gemacht.

	real x0=x;		// x0 mu verndert werden. Spter wird zuerst aus
		// x0 die Quadratwurzel gezogen und daraus dann die Wurzel aus x berechnet.
	if(x.negativ)
	{
		x0.negativ=false;
		GleitkommaRunden=-GleitkommaRunden;		// wird unten rckgngig gemacht.
	}
	int16 add_res_exp2;
	if(x.exp2 >= 0)
	{
		x0.exp2=x.exp2 & 1;			// Nun ist 1 <= x0 < 4 .
		add_res_exp2=x.exp2 / 2;
	}
	else
	{
		x0.exp2=((-x.exp2) & 1);	// Nun ist 1 <= x0 < 4 (wie im Falle x.exp2 >= 0).
		add_res_exp2= - ((1 - x.exp2) / 2);
	}

	real y(real::Anz16bits(x0.anz_mts + 1));

	if(x0.anz_mts<3)
		x0.ResizeMantisseAnz16Bits(3);
		// WICHTIG: im folgenden wird vorausgesetzt, da x0 mindestens 3 Mantisse-
		// worte hat.

	y.ResizeMantisseAnz16Bits(1);		// Der weitere Speicher bleibt zugewiesen; deshalb
		// mu bei einer folgenden schrittweisen Erhhung der Genauigkeit der
		// Mantisse von y nicht immer der bisherige Speicher freigegeben und neuer
		// zugewiesen werden.
		// ACHTUNG: ResizeMantisseAnz16Bits() darf nicht zwischen den unten stehenden
		// Anweisungen, welche die Mantisse und den Exponenten von y verndern,
		// stehen, weil ResizeMantisseAnz16Bits() immer eine korrekt initialisierte Zahl
		// voraussetzt; insbesondere fhrt ResizeMantisseAnz16Bits beim Verkleinern
		// der Mantissenlnge ggf. Rundungen durch.

	uint16 d;
	if(1==x0.exp2)		// andernfalls kann x0.exp2 nur gleich Null sein.
		d=real::integer_sqrt((((unsigned long)x0.mts[0])<<16LU) + ((unsigned long)x0.mts[1]));
	else
		d=real::integer_sqrt(((((unsigned long)x0.mts[0])<<16LU) + ((unsigned long)x0.mts[1])) >> 1LU);
	y.mts[0]=d;
	y.null=y.negativ=false;		// Die Mantisse wird im folgenden zugewiesen.
	y.exp2=0;

	// y ist nun (bis auf den zu addierenden Exponenten add_res_exp2)
	// mit einer Genauigkeit von  2 hoch -40  gleich der Wurzel aus x.
	// Bei dieser Aussage wurden sicherheitshalber etwa 8 hintere Bits
	// nicht bercksichtigt.

	// Wegen der obigen Initialisierung gilt nun y.negativ==false und
	//    y.exp2 == 0 ,
	// also  1 <= y < 2 .

	int16 genau=10, diff, mask;		// mu mindestens gleich 2 sein, damit die folgende
		// Schleife keine Endlosschleife ist.
		// Die Variable y nhert die Wurzel aus x0 mit der Genauigkeit
		//  2 hoch (-genau)  an.

	real x0_kurz(real::Anz16bits(x0.anz_mts));
	int16 soll_genau=16*x0.anz_mts + 4;
	bool first=true;

	while(genau < soll_genau)		// Newton-Iteration
	{
		//printf("genau=%d   soll_genau=%d\n", genau, soll_genau);
		genau=2*genau-1;		// *2 : Verdoppelung der Genauigkeit bei jedem
			// Interationsschritt, wenn dieser aus den Ausgangsdaten (die die
			// Genauigkeit  2 hoch -'genau'  haben) durch Rechenschritte mit
			// beliebig hoher Genauigkeit ausgefhrt wrden.
			// Durch die Verkleinerung von 'genau' um 1 werden Rundungsfehler in der Iteration bercksichtigt.

		if(first)
		{
			int16 g=soll_genau;
			while(g>genau)
				g=(g/2)+1;
			genau=g;
			for(mask=1 ; g<soll_genau; g=2*g-1)
				mask <<= 1;
			diff=g-soll_genau;
			first=false;
		}
		else if(0!=(diff & (mask>>=1)))
			--genau;

		// if(genau > soll_genau+1) genau=soll_genau+1; nicht ntig !!!

		int16 anz16b=(genau + (4+15) )/16;	// Dies bewirkt, da die Rechnung mit
			// mindestens  'genau'+5  Mantissebits durchgefhrt wird, WOBEI
			// MIT  'genau'  DER WERT DIESER VARIABLEN GEMEINT IST, DER AN
			// DIESER PROGRAMMSTELLE, ALSO VOR DEM '--genau' ANGENOMMEN WIRD.

		x0_kurz=x0;
		x0_kurz.ResizeMantisseAnz16Bits(anz16b);
			// x0_kurz statt x0 verwenden: Dies spart Rechenzeit bei den ersten
			// Iterationen. Bei Zahlen mit hoher Genauigkeit (z.B. 16*10 Mantissebits)
			// kann man sagen, da dann die Funktion sqrt() nherungsweise weniger
			// Rechenzeit bentigt als man fr zwei Divisionen MIT DER VOLLEN
			// MANTISSENLNGE, also mit der Mantissenlnge des Arguments 'x',
			// bentigt.

 		y = (y + (x0_kurz/y)).mult2hochX(-1);
	};
	//printf("genau=%d   soll_genau=%d\n", genau, soll_genau);

	if(x.negativ)
	{
		y.negativ=true;  // Eigentlich ist eine Wurzel aus einer negativen Zahl
							  // imaginr, aber hier soll eine C++-Ausnahme vermieden werden.
	}

	GleitkommaRunden=runden_vorher;  // vor dem RoundUpDown-Befehl !!!
	y.RoundUpDown(16*x0.anz_mts+4);
		// Nun ist noch eine Genauigkeit von  2 hoch -(16*x0.anz_mts+2)
		// sichergestellt.

	y.mult2hochX(add_res_exp2);
	y.ResizeMantisseAnz16Bits(x.anz_mts);

	return y;
}

//////////////////////////////////////////////////////////////////////////
