//***********************************************************************************************************//
//***********************************************************************************************************//
//                                                                                                           //
// Gleitkomma.cpp and Gleitkomma.h : Used for definition of the class real. An Object of type real           //
//      represents a floating point number with adjustable precision (number of mantissa bits/words).         //
//                                                                                                           //
// Autor: Florian Knigstein, mail@virgusta.eu                                                               //
//                                                                                                           //
// Compiler: avr-gcc                                                                                         //
//                                                                                                           //
// Copyright:                                                                                                //
//      You may use this program free of charge for any purpose you wish provided you agree to the following //
//      License:                                                                                             //
//      Warranty of Provenance and Disclaimer of Warranty. Licensor warrants that the copyright              //
//      in and to the Original Work and the patent rights granted herein by Licensor are owned by the        //
//      Licensor or are sublicensed to You under the terms of this License with the permission of the        //
//      contributor(s) of those copyrights and patent rights. Except as expressly stated in the immediately  //
//      preceding sentence, the Original Work is provided under this License on an "AS IS" BASIS and WITHOUT //
//      WARRANTY, either express or implied, including, without limitation, the warranties of                //
//      non-infringement, merchantability or fitness for a particular purpose. THE ENTIRE RISK AS TO THE     //
//      QUALITY OF THE ORIGINAL WORK IS WITH YOU. This DISCLAIMER OF WARRANTY constitutes an essential part  //
//      of this License. No license to the Original Work is granted by this License except under this        //
//      disclaimer.                                                                                          //
//                                                                                                           //
//      Limitation of Liability. Under no circumstances and under no legal theory, whether in tort           //
//      (including negligence), contract, or otherwise, shall the Licensor be liable to anyone for any       //
//      indirect, special, incidental, or consequential damages of any character arising as a result of      //
//      this License or the use of the Original Work including, without limitation, damages for loss of      //
//      goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages    //
//      or losses. This limitation of liability shall not apply to the extent applicable law prohibits such  //
//      limitation.                                                                                          //
//                                                                                                           //
//***********************************************************************************************************//
//***********************************************************************************************************//


//***** ATTENTION ******************************************************************************************//
// This library makes use of dynamic memory allocation. But malloc's return value is NOT verified to be     //
// non zero, i.e. "out of memory" will not be handled properly. The program will likely crash if out of     //
// memory occurs. For workaround you might modify the program and generate a NaN (not an number) if malloc  //
// returns zero.                                                                                            //
//**********************************************************************************************************//


#ifndef __GLEITKOMMA_INCLUDED
#define __GLEITKOMMA_INCLUDED


#define ANZAHL_DEZIMALE_MANTISSESTELLEN 50
// #define ENABLE_GLEITKOMMA_EXCEPTIONS

// Wenn im Folgenden DIVISION_DURCH_NEWTON_ITERATION definiert wird, wird zur Division ein Verfahren
// angewandt, welches zwar etwa 1500 Bytes weniger Programmspeicher bentigt, dafr aber eine lngere
// Ausfhrungszeit bentigt.
#define DIVISION_DURCH_NEWTON_ITERATION

#define ONLY_NAN_NO_INFINITY


#define NDEBUG

#include "stdlib.h"
#include "string.h"

typedef signed char int8;
typedef signed int int16;
typedef unsigned char uint8;
typedef unsigned int uint16;

extern int8 StdZahlMantissewoerter;

extern int8 GleitkommaRunden;
		// Wenn GleitkommaRunden > 0 gilt, so wird bei Bedarf immer aufgerundet.
		// Bei negativen Zahlen verkleinert sich in einem solchen Fall der Betrag.
		// Jede der vier Grundrechenarten sowie der Copy-Konstruktor mit Verringerung
		// der Genauigkeit liefern niemals eine Zahl, die kleiner ist als die
		// genaue Zahl, die sich ergeben wrde, wenn man die Eingangsdaten als genau
		// betrachtet. Entsprechendes gilt fr die Umwandlung von Zahlen in der
		// Dezimaldarstellung (Strings) in realen sowie fr die Umwandlung
		// in umgekehrter Richtung.
		// Wenn GleitkommaRunden < 0 gilt, so wird bei Bedarf abgerundet.
		// Bei negativen Zahlen vergrert sich in einem solchen Fall der Betrag.
		// Wenn GleitkommaRunden==0, so wird auf die nchste darstellbare Zahl
		// gerundet.

class real
{
private:
		// ---- Achtung: ---- Der operator=() mu ALLE Variablen sowie das Feld mts[]
		// kopieren !!!
	bool null, negativ, unendlich;
		// Wenn null==true UND unendlich==true gilt,
		// wird das Objekt "Not a number" dargestellt.
		// Falls jedoch #define ONLY_NAN_NO_INFINITY definiert wird, wird NaN durch unendlich=true und
		// einen beliebigen Wert von null dargestellt.

	int16 exp2;		// nur gltig, wenn null==false ist !
	static int16 min_exp2;	// min_exp2 = - (1u << 13)  (einschlielich)
	static int16 max_exp2;	// max_exp2 =   (1u << 13)  (einschlielich)
	typedef unsigned short digit;
	enum { bits_per_digit=16 }; // Dieser Wert darf nicht gendert werden, weil die Programmteile
		// der Klasse real sich nicht automatisch an einen anderen Wert anpassen !

	int8 anz_mts;	// Die Zahl der UWORDs, die das Feld mts[] (MINDESTENS) beinhaltet.
			// Es wird immer angenommen, da anz_mts >= 1 gilt und entsprechend Speicher
			// an mts zugewiesen wurde. Wenn null==false gilt, so werden die Daten,
			// die im mts[0] bis mts[anz_mts-1] stehen, als initialisiert betrachtet.
			// -- ACHTUNG --: Verschiedene Funktionen setzen voraus, da z.B. der
			// Zuweisungsoperator  operator=()  der Variablen anz_mts immer genau den
			// Wert zuweist, den anz_mts der zuzuweisenden real hat.
	int8 AnzMantisseMemory;	// AnzMantisseMemory gibt die Zahl der UWORDs an, die
			// tatschlich fr mts[] reserviert worden sind.
			// Es gilt immer  AnzMantisseMemory >= anz_mts >= 1 .
			// Manche Funktionen bentigen temporr eine hhere Genauigkeit, um ihr
			// Ergebnis so genau berechnen zu knnen, da 'anz_mts' Mantissestellen
			// bis auf eine eventuelle Abweichung von z.B.  +-ZWEI  der letzten
			// dieser Stellen genau ist.

	digit *mts;		// Mantisse. ACHTUNG: Wenn null==true gilt, so ist mts[]
						//					nicht initialisiert !
		// Es wird die Zahl
		//		( mts[0] + mts[1]/65536.0 + mts[2]/(65536.0 hoch 2) +  ...
		//					+ mts[anz_mts-1]/(65536.0 hoch (anz_mts-1)
		//    )
		//    mal ( 2 hoch (exp2-15) )
		// dargestellt, wenn null==false und negativ==false ist.
		// Wenn die real korrekt initialisiert ist und null==false ist,
		//			so gilt immer  mts[0]>=32768u (Normalisierung) .
		// ---- WICHTIG ----: unsigned (UWORD) fr die Mantisse. Denn der Mantisse
		// Wertebereich von 0 bis 65535 wird ausgenutzt.
	
	void init(int8 anz_mts_16bits);
	void normalisiere();
	void check_exp();
	void UebertragUndRunden(int8 runden, bool uebertrag, unsigned short weitere_stellen);
	
		// Achtung: Die folgenden vier Funktionen liefern Referenzen auf eine
		// statische Variable der betreffenden Funktion !
	friend real& addiere_betraege(const real& x, const real& y,
											int8 runden);
	friend real& subtrahiere_betraege(const real& x, const real& y,
												int8 runden);
	friend real& multipliziere(const real& x, const real& y);
#ifdef DIVISION_DURCH_NEWTON_ITERATION
	friend real& dividiere_newton_iterationen(const real& x, const real& y, real* rest=0);
#else
	friend real& dividiere(const real& x, const real& y, real **rest = 0);
#endif

	private: real(const double d) { }
			// Es ist wichtig, da auch ein Konstruktor von double-Zahlen existiert.
			// Wenn nur ein Konstruktor aus long-Werten existieren wrde, so
			// wrde der Code
			//    real z = 5.77;
			// die ganze Zahl 5 erzeugen, denn es wrde zuerst eine Konversion von
			// double (5.77) nach integer (5) stattfinden, um dann damit den Konstruktor real(long) aufzurufen.
			// ****** Aus Speichergrnden (Mikrokontroller mit begrenztem Flash-Speicher) wird der Konstruktor
			// von double jedoch nicht implementiert. Es ist wichtig, dass statt dessen als private definiert wird,
			// damit der Benutzer beim Versuch, z.B. real(5.77) zu erzeugen, eine Fehlermeldung erhlt.

	void naechsthoehererBetrag();
	void naechstkleinererBetrag();

public:
	class OverflowError { };
	class DivisionByZeroError { };

	struct AnzBits
	{
		int16 anz_bits;
		explicit AnzBits(int16 _anz_bits) : anz_bits(_anz_bits) { };
	};
	struct AbsLbGenau
	{
		int16 minus_lb_rel_error;
		explicit AbsLbGenau(int16 _minus_lb_rel_error) : minus_lb_rel_error(_minus_lb_rel_error) { };
	};
	struct Anz16bits
	{
		int8 anz;
		explicit Anz16bits(int8 Anz) : anz(Anz) { };
		Anz16bits(const AnzBits& anzb) : anz((anzb.anz_bits+bits_per_digit-1)/bits_per_digit) { };
		Anz16bits(const AbsLbGenau& alg) : anz((alg.minus_lb_rel_error+bits_per_digit-1)/bits_per_digit) { };
	};
	real(Anz16bits anz=Anz16bits(StdZahlMantissewoerter))
		{ init(anz.anz); }	// mit Null initialisieren.
	real(const real& x);	// x mu eine Referenz sein, sonst findet eine unendliche Rekursion statt !!!
			// Copy-Konstruktor, bernahme der Genauigkeit der bergebenen Zahl
	real(const real& x, Anz16bits anz);
			// Copy-Konstruktor, Redefinition der Genauigkeit. Eventuell Auffllen
			// der Mantisse mit Nullen am Ende oder aber Abschneiden und Runden.
	real(const char *wert,  Anz16bits anz=Anz16bits(StdZahlMantissewoerter));
			// 'wert' mu die Dezimaldarstellung der Zahl sein, mit der initialisiert werden soll.
	real(int16 Exp2, int8 Anz_mts, unsigned short *Mantisse, bool negativ=false);
			// Hier mu 'Mantisse' auf 'Anz_mts' UWORDs zeigen.
	inline real(const int N, Anz16bits anz=Anz16bits(StdZahlMantissewoerter))  { initialize_with_long(N, anz); }
	inline real(const long N, Anz16bits anz=Anz16bits(StdZahlMantissewoerter)) { initialize_with_long(N, anz); }
	inline real(const unsigned int N, Anz16bits anz=Anz16bits(StdZahlMantissewoerter))  { initialize_with_ulong(N, anz); }
	inline real(const unsigned long N, Anz16bits anz=Anz16bits(StdZahlMantissewoerter)) { initialize_with_ulong(N, anz); }

	//explicit real(const unsigned long *digits, int8 __anz_digits, int8 __bits_per_digit, int16 __exp2=0,
	//						Anz16bits anz=Anz16bits(0));
	// Achtung: digits[0] ist das niederwertigste Wort und digits[__anz_digits-1] das hchstwertigste
	// -- im Gegensatz zur Organisation der Worte in real::mts[] . Auerdem wird die in digits[]
	// abgespeicherte Zahl als GANZE Zahl interpretiert, welche mit 2^__exp2 multipliziert wird.

	real& operator=(const real& x);
	~real();

#ifndef ONLY_NAN_NO_INFINITY
	static real MakePlusInfinity()
		{ real inf(Anz16bits(1)); inf.unendlich=true;
			inf.negativ=false; inf.null=false; return inf; }
	static real MakeMinusInfinity()
		{ real inf(Anz16bits(1)); inf.unendlich=true;
			inf.negativ=true; inf.null=false; return inf; }
#endif
	static real MakeNaN()
		{ real z(Anz16bits(1)); z.unendlich=true;
			z.null=true; return z; }

			// Die Funktion ResizeMantisseAnz16Bits() darf nur dann innerhalb eines Konstruktors
			// von class GleitkommmaZahl aufgerufen werden, wenn die real
			// schon eine gltige Initialisierung besitzt.
	real& ResizeMantisseAnz16Bits(int8 anz);
			// Nach ResizeMantisseAnz16Bits gilt anz_mts==anz, d.h. anz ist die neue Zahl
			// der zu verwendenden Mantisseworte. Es sei anz_mts_alt der Wert von
			// anz_mts VOR dem Aufruf von ResizeMantisseAnz16Bits(). Wenn nun  anz < anz_mts_alt
			// bergeben wird, so wird entsprechend  'GleitkommaRunden'  gerundet,
			// sofern null==false ist. Der an mts[] zugewiesene Speicher wird
			// bei  anz < anz_mts_alt  NICHT teilweise freigegeben, und
			// AnzMantisseMemory behlt seinen Wert. Wenn aber  anz > anz_mts_alt
			// bergeben wird, so wird anz_mts angepat, bei Bedarf ein grerer
			// Speicherbereich fr mts[] reserviert und AnzMantisseMemory entsprechend
			// gesetzt. Wenn dann  null==false  gilt, so werden die hinzugekommenen
			// Mantisseworte mit Null gefllt.
			// Es wird danach *this zurckgeliefert.
	real& ResizeMantisse(const AnzBits& anz_bits)
		{ return ResizeMantisseAnz16Bits((anz_bits.anz_bits+15)/16); }

	inline int16& get_exp2() { return exp2; }
	inline unsigned short *GetMantissePtr() const { return mts; }
	static inline int8 get_bits_per_digit() { return bits_per_digit; }
	int8 GetAnzMts() const { return anz_mts; }

#ifdef ONLY_NAN_NO_INFINITY
	bool is_NaN() const { return true==unendlich; }
#else
	bool is_NaN() const { return true==null && true==unendlich; }
		// Wenn null==true UND unendlich==true gilt,
		// wird das Objekt "Not a number" dargestellt.
	bool is_plus_infinity() const { return true==unendlich && false==null && false==negativ; }
	bool is_minus_infinity() const { return true==unendlich && false==null && true==negativ; }
	bool is_infinite() const { return true==unendlich && false==null; }
			// ACHTUNG: is_infinite() liefert nicht immer das logische Komplement
			// zu is_finite(), denn bei NaN liefert is_infinite() den Wert false.
	inline bool is_finite() const { return false==unendlich; }	
#endif
	bool is_negative() const { return true==negativ && false==null; }
	bool is_positive() const { return false==negativ && false==null; }
	bool is_not_negative() const { return (true==null && false==unendlich)
									|| (false==null && false==negativ); }
	bool is_not_positive() const { return (true==null && false==unendlich)
									|| (false==null && true==negativ); }
	bool is_zero() const { return true==null && false==unendlich; }
	int16 GetLsbNumber() const;

	friend real operator+(const real& x, const real& y);
	friend real operator-(const real& x, const real& y);
	friend real operator*(const real& x, const real& y);
	friend real operator/(const real& x, const real& y);
	friend real operator-(const real& x);	// einstelliges Minus
					// einstelliges Plus ist auch definiert, aber kein friend notwendig.

	real& operator+=(const real& y);
	real& operator-=(const real& y);
	real& operator*=(const real& y);
	real& operator/=(const real& y);

	friend real fabs(const real& x);
	friend real sqrt(const real& x);
	friend bool operator==(const real& x, const real& y);
				// operator!=() mu kein friend sein.
	friend bool operator> (const real& x, const real& y);
	friend bool operator< (const real& x, const real& y);
				// operator>=() und operator<=() mssen keine friends sein.
	
	static real MinMaxPositiveZahl(Anz16bits anz, bool max);
	static inline real MinPositiveZahl(Anz16bits anz) { return MinMaxPositiveZahl(anz, false); }
	static inline real MaxPositiveReelleZahl(Anz16bits anz) { return MinMaxPositiveZahl(anz, true); }
	static real _10HochX(int16 x, int16 AbsLbGenau, int16 MinAnzMtsBits=0);
	// friend real NaechstGroessereGltZahl(const real& x);
	// friend real NaechstKleinereGltZahl(const real& x);
	real& RoundUpDown(int16 genau_vorher);
	real& mult2hochX(int16 x);
	friend bool vergleiche_betraege(const real& x, const real& y);
	friend long to_long(const real& x);
	friend real modf(const real& x, real *integer = 0);
	friend real fmod(const real& x, const real& y, real *ganz = 0);

	friend real to_integer(const real& x);
	// friend bool is_integer(const real &x);
	// friend real floor(const real &x);
	// friend real ceil(const real &x);
	friend inline bool _finite(const real& x);
	friend inline bool _isnan(const real& x);

	int8 AnzMtsBenutzt() const;
	static int8 AnzBitsBenutztInLong(long x);

	char *to_decimalExp(int8 anz_dezimal_mantisse, bool MantisseUndExponentGetrennt=false,
						int16 *ExponentBasis10=0) const;
	char *optimal_decimal_representation(uint8 max_nr_chars, uint8 max_leading_mantisse_zeros=3);
	
	static unsigned short integer_sqrt(unsigned long x)
	{	// Berechnet die nchste ganze Zahl, die nicht grer als die Wurzel aus x ist.
		unsigned long t=1LU<<30LU, r=0, s;
		for(unsigned char i=17; 0!=--i; )
		{
			s = t + r;
			r>>= 1LU; 
			if(s <= x)
				{ x -= s; r |= t;}
			t >>= 2LU;
		}
		return (unsigned short)r;
	}
	
	static inline real number_zero()
	{
		return real(Anz16bits(1));
	}
	static real number_one()
	{
		return real(1L, Anz16bits(1));
	}
	static real number_ten()
	{
		return real(10L, Anz16bits(1));
	}
private:
	void initialize_with_long(const long N, Anz16bits anz=Anz16bits(StdZahlMantissewoerter));
	void initialize_with_ulong(const unsigned long N, Anz16bits anz=Anz16bits(StdZahlMantissewoerter));
};

#ifndef ONLY_NAN_NO_INFINITY
inline bool _finite(const real& x)
{
	return false==x.unendlich;
}
inline bool _isnan(const real& x)
{
	return true==x.unendlich && false==x.null;
}
#else
inline bool _isnan(const real& x)
{
	return true==x.unendlich;
}
#endif


bool operator>=(const real& x, const real& y);
bool operator<=(const real& x, const real& y);

long to_long(const real& x);
		// Die Nachkommastellen von x werden nicht beachtet, d.h. es wird z.B.
		//  4.1  zu  (long)4  und  -4.1  zu  (long)(-4) .
		// --- ACHTUNG ---: Wenn der so entstehende ganzzahlige Teil von x
		// auerhalb des Bereichs ist, der durch long-Zahlen dargestellt
		// werden kann, so ist das Ergebnis undefiniert !

// real& dividiere(const real& x, const real& y, real **rest);
real& dividiere_newton_iterationen(const real& x, const real& y,
											 real* rest/*=0*/);

real modf(const real& x, real *integer);

real to_integer(const real& x);
			// Liefert die ganze real, die aus x durch Abschneiden der
			// Kommastellen entsteht.

//real NaechstGroessereGltZahl(const real& x);
//real NaechstKleinereGltZahl(const real& x);

real operator+(const real& x, const real& y);
real operator-(const real& x, const real& y);
real operator-(const real& x);	// einstelliges Minus
		// Achtung: Wert und keine Referenz zurckliefern !
//real operator+(const real& x);	// einstelliges Plus
real fabs(const real& x);
bool operator!=(const real& x, const real& y);
bool operator==(const real& x, const real& y);
bool operator>=(const real& x, const real& y);
bool operator<=(const real& x, const real& y);
bool operator>(const real& x, const real& y);
bool operator<(const real& x, const real& y);
bool vergleiche_betraege(const real& x, const real& y);
real operator*(const real& x, const real& y);
real operator/(const real& x, const real& y);
real fmod(const real& x, const real& y,
						  real *ganz);
		// ..... Default fr ganz ist der Null-Zeiger.
		// ----- ACHTUNG -----: DIESE FUNKTION ARBEITET NUR KORREKT, WENN GILT:
		//     abs(x) <= abs(y) * (2 hoch (max_anz*16))
		// Dabei ist 'max_anz' das Maximum der Anzahlen der Mantissewrter
		// der Zahlen x und y (jedes Mantissewort hat 16 Bits).
		// Auerdem mu  y != 0  gelten.
		// -------------------
		// Wenn diese Bedingungen erfllt sind, so werden eine ganze Zahl  n  und
		// eine real  f  berechnet, so da
		//      x == n*y + f    und     0 <= abs(f) < y
		// gelten und  f  das gleiche Vorzeichen hat wie x.
		// Die Zahl  f  wird von fmod() zurckgeliefert. Wenn  ganz != 0  bergeben
		// wird, so wird nach  *ganz  die ganze Zahl  n  geschrieben.
		// ===> Es spielt fr die Ergebnisse von fmod() keine Rolle, welchen
		//      Wert die Variable 'GleitkommaRunden' vor dem Aufruf von fmod()
		//      hatte. Natrlich hat 'GleitkommaRunden' nach dem Aufruf von
		//      fmod() den gleichen Wert wie vor dem Aufruf.
		//
		// ===> Wenn obige Bedingungen an x und y erfllt sind, SO GILT DIE GLEICHUNG
		//        x == n*y + f  EXAKT, WENN MAN VORAUSSETZT, DASS DIE
		//      RECHNUNG  n*y + f  MIT MINDESTENS  max_anz + y.anz_mts
		//      Mantissewrtern zu je 16 Bits durchgefhrt
		// wird, wobei  max_anz  das Maximum der Zahl der Mantissewrter
		// von x und y sei und weiter  y.anz_mts  die Zahl der Mantissewrter von y.
		// Das heit, vor der Durchfhrung der Rechnung  n*y + f  mte man die
		// von fmod() gelieferten Werte  n  und  f  auf  max_anz + y.anz_mts  Mantisse-
		// wrter erweitern (mit ResizeMantisseAnz16Bits()), denn:
		//
		// Sowohl der Rckgabewert f als auch die eventuell nach  *ganz  geschriebene
		// Zahl  n  besitzen soviele Mantissewrter wie das Maximum der Anzahl
		// der Mantissewrter der Zahlen x und y.

real sqrt(const real& x);
		// Wenn x positiv ist, so wird ein positives y geliefert, fr das
		// (nherungsweise) y*y = x gilt.
		// Wenn x negativ ist, so wird ein negatives y geliefert, fr das
		// (nherungsweise) y*y = -x gilt.
		// Wenn man den bergebenen Wert x als exakt betrachtet, so wird
		// durch sqrt(x) die Quadratwurzel von x mit einer Genauigkeit
		// angenhert, die nicht schlechter ist als
		//   2 hoch -(16*x.anz_mts - 2) .
		// Die Variable GleitkommaRunden wird beachtet.

// double to_double(const real& x);


///////////////////////////////////////////////////////////////////////////////////////////////////////////////



#endif  //  __GLEITKOMMA_INCLUDED
