/************************************************************************/
/*                                                                      */
/*                      RC5 Remote Receiver                             */
/*                                                                      */
/*              Author: Peter Dannegger                                 */
/*                      danni@specs.de                                  */
/*                                                                      */
/************************************************************************/

// Anpassungen vorgenommen von Florian Knigstein

#include "rc5.h"

#define RC5TIME 	1.778e-3		// 1.778msec
#define PULSE_MIN	(uchar)(XTAL_RC5 / 512 * RC5TIME * 0.4 + 0.5)
#define PULSE_1_2	(uchar)(XTAL_RC5 / 512 * RC5TIME * 0.8 + 0.5)
#define PULSE_MAX	(uchar)(XTAL_RC5 / 512 * RC5TIME * 1.2 + 0.5)

#define DDR(x) (*(&x - 1))      /* address of data direction register of port x */

uchar	rc5_bit=0;				// bit value
uchar	rc5_time=0;				// count bit time
uint	rc5_tmp=0;				// shift bits in
uint	rc5_data=0;				// store result

void rc5_init(void)
{
	DDR(xRC5_PORT) &= ~(1<<xRC5_PINNR);
	xRC5_PORT |= (1<<xRC5_PINNR);
	TCCR0 = 1<<CS02;			//divide by 256
	TIMSK = 1<<TOIE0;			//enable timer interrupt
	sei();
}

extern void usart_write_str(char *buffer);

uint read_rc5(void)
{
	static uint last_i=0;
	cli();
	uint i = rc5_data;			// read two bytes from interrupt !
	rc5_data = 0;
	sei();
	if(0==i)
		return 0;
	else if(last_i!=i) // Nach dem Reset ist diese Bedingung beim ersten Tastendruck erfllt.
	{		// Spter ist sie genau dann erfllt, wenn der Tastencode wegen lngerem Gedrckthalten
			//  einer Taste wiederholt gesandt wird (nicht zum ersten Mal nach dem Drcken der Taste).
		last_i=i;
		return i &~ (1<<11);
	}
	else
		return i | (1<<11);
}

SIGNAL (SIG_OVERFLOW0)
{
  uint tmp = rc5_tmp;				// for faster access

  TCNT0 = -2;					// 2 * 256 = 512 cycle

  if( ++rc5_time > PULSE_MAX ){			// count pulse time
    if( !(tmp & 0x4000) && tmp & 0x2000 )	// only if 14 bits received
      rc5_data = tmp;
    tmp = 0;
  }

  if( (rc5_bit ^ xRC5_IN) & 1<<xRC5 ){		// change detect
    rc5_bit = ~rc5_bit;				// 0x00 -> 0xFF -> 0x00

    if( rc5_time < PULSE_MIN )			// to short
      tmp = 0;

    if( !tmp || rc5_time > PULSE_1_2 ){		// start or long pulse time
      if( !(tmp & 0x4000) )			// not to many bits
        tmp <<= 1;				// shift
      if( !(rc5_bit & 1<<xRC5) )		// inverted bit
        tmp |= 1;				// insert new bit
      rc5_time = 0;				// count next pulse time
    }
  }

  rc5_tmp = tmp;
}
